"""
SQLStore to support Kypher queries over KGTK graphs.
"""

import sys
import os
import os.path
import sqlite3
import time
import csv
import io
import re
from   functools import lru_cache
import itertools
import copy

import sh

from   kgtk.exceptions import KGTKException
from   kgtk.kypher.utils import *
from   kgtk.kypher.kgtkinfo import KgtkInfoTable
import kgtk.kypher.indexspec as ispec
from   kgtk.kypher.functions import SqlFunction


### TO DO:

# o automatically run ANALYZE on tables and indexes when they get created
#   - we decided to only do this for indexes for now
# - support naming of graphs which would allow deleting of the source data
#   as well as graphs fed in from stdin
# + absolute file names are an issue when distributing the store
# - support some minimal sanity checking such as empty files, etc.
# - handle column name dealiasing and normalization
# o explanation runs outside the sqlite connection and thus does not see
#   user functions such as kgtk_stringify and friends which causes errors;
#   fixed for --explain
# - support declaring and dropping of (temporary) graphs that are only used
#   once or a few times
# - allow in-memory graphs, or better, support memory-mapped IO via
#   PRAGMA mmap_size=NNN bytes, which would be transparent and usable on demand
# o support other DB maintenance ops such as drop, list, info, etc.
# - check for version of sqlite3, since older versions do not support ascii mode
# - protect graph data import from failure or aborts through transactions
# - handle table/index creation locking when we might have parallel invocations,
#   but it looks like sqlite already does that for us
# - provide some symbolic graph size classification (small/medium/large/xlarge)
#   and implement table optimizations based on those categories
# - support bump_timestamp or similar to better keep track of what's been used
# - improve table definitions to define core columns as required to be not null
# - full LRU cache maintainance, but maybe abandon the whole LRU concept and
#   call it a store and not a cache
# + complete literal accessor functions
# + handle VACUUM and/or AUTO_VACUUM when graph tables get deleted
#   - actually no, that requires a lot of extra space, so require to do that manually


### SQL Store

class SqlStore(object):
    """SQL database capable of storing one or more KGTK graph files as individual tables
    and allowing them to be queried with SQL statements.
    """
    # This is just an abstract place-holder for now.  Once we complete SqliteStore
    # and generalize this to other SQL DB(s), we'll move API-level methods up here.
    pass


class SqliteStore(SqlStore):
    """SQL store implemented via sqlite3 (which is supported as a Python builtin library.
    """

    MASTER_TABLE = sdict[
        '_name_': 'sqlite_master',
        'columns': sdict[
            # not sure about the real types of these, but that shouldn't matter:
            'type':     sdict['_name_': 'type',     'type': 'TEXT'],
            'name':     sdict['_name_': 'name',     'type': 'TEXT'],
            'tbl_name': sdict['_name_': 'tbl_name', 'type': 'TEXT'],
            'rootpage': sdict['_name_': 'rootpage', 'type': 'INTEGER'],
            'sql':      sdict['_name_': 'sql',      'type': 'TEXT'],
        ]
    ]

    # Files contain KGTK data defining graphs, and graphs are SQL tables representing that data.
    # They are represented as separate object types, but for now the association is 1-1 where each
    # file points to the graph it defines and each graph is named by its associated file.
    # However, in the future we might redefine this association, e.g., multiple files could define
    # a graph, in which case graphs should have their own external names.  The latest implementation
    # of INFO_TABLE as a schema-free KGTK table makes it easy to add new kinds of information over
    # time without invalidating the information stored in existing graph caches.

    INFO_TABLE = sdict[
        '_name_': 'kypher_master',
        'columns': sdict[
            'node1': sdict['_name_': 'node1', 'type': 'TEXT'],
            'label': sdict['_name_': 'label', 'type': 'TEXT'],
            'node2': sdict['_name_': 'node2', 'type': 'TEXT'],
            'id':    sdict['_name_': 'id',    'type': 'TEXT'],
        ],
        'without_rowid': False, # just for illustration
        'temporary': False,     # just for illustration
    ]

    def __init__(self, dbfile=None, create=False, loglevel=0, conn=None, readonly=False):
        """Open or create an SQLStore on the provided database file 'dbfile'
        or SQLite connection object 'conn'.  If 'dbfile' is provided and does
        not yet exist, it will only be created if 'create' is True.  Passing
        in a connection object directly provides more flexibility with creation
        options.  In that case any 'dbfile' value will be ignored and instead
        looked up directly from 'conn'.
        """
        self.loglevel = loglevel
        self.dbfile = dbfile
        self.conn = conn
        self.readonly = readonly
        if not isinstance(self.conn, sqlite3.Connection):
            if self.conn is not None:
                raise KGTKException('invalid sqlite connection object: %s' % self.conn)
            if self.dbfile is None:
                raise KGTKException('no sqlite DB file or connection object provided')
            if not os.path.exists(self.dbfile) and (not create or readonly):
                raise KGTKException('sqlite DB file does not exist: %s' % self.dbfile)
        else:
            self.dbfile = self.pragma('database_list')[0][2]
        self.user_functions = set()
        self.vector_store = None
        self.init_meta_tables()
        self.configure()
        # run this right after 'configure()' since it is not thread
        # safe and later calls might lead to undefined behavior;
        # also we might need it for distinct or order-by queries:
        self.configure_temp_dir()

    def log(self, level, message):
        if self.loglevel >= level:
            header = '[%s sqlstore]:' % time.strftime('%Y-%m-%d %H:%M:%S')
            sys.stderr.write('%s %s\n' % (header, message))
            sys.stderr.flush()

    def init_meta_tables(self):
        self.infotable = InfoTable(self, self.INFO_TABLE._name_)

    def describe_meta_tables(self, out=sys.stderr):
        """Describe the current content of the internal bookkeeping tables to 'out'.
        """
        out.write('Graph Cache:\n')
        out.write('DB file: %s\n' % self.dbfile)
        out.write('  size:  %s' % format_memory_size(self.get_db_size()))
        out.write('   \tfree:  %s' % format_memory_size(self.get_db_free_size()))
        out.write('   \tmodified:  %s\n' % time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(os.path.getmtime(self.dbfile))))
        out.write('\n')
        out.write('KGTK File Information:\n')
        self.describe_file_info_table(out=out)
        out.write('\n')
        out.write('Graph Table Information:\n')
        self.describe_graph_info_table(out=out)

    # TO DO: consider reducing this or making it configurable, since its effect on runtime
    #        seems to be small (5-10%) compared to the memory it additionally consumes:
    CACHE_SIZE = 2 ** 32            #  4GB
    #CACHE_SIZE = 2 ** 34           # 16GB
    #CACHE_SIZE = 2 ** 34 + 2 ** 33 # 24GB

    def configure(self):
        """Configure various settings of the store.
        """
        #self.pragma('main.page_size = 65536') # for zfs only
        self.pragma('main.cache_size = %d' % int(self.CACHE_SIZE / self.pragma('page_size')))

    def configure_temp_dir(self):
        """Configure the SQLite temp directory to be in the same location as the database file,
        unless that is explicitly overridden by a different settings from SQLITE_TMPDIR.
        This tries to avoid disk-full errors when large files are imported and indexed, since
        standard temp directories are usually located in smaller OS disk partitions.
        """
        # 'temp_store_directory' is deprecated and might not be available, but there is no good
        # alternative except SQLITE_TMPDIR, so we try anyway and report just a warning if --debug:
        try:
            if not self.pragma('temp_store_directory') and not os.getenv('SQLITE_TMPDIR') and self.dbfile:
                tmpdir = os.path.dirname(os.path.realpath(self.dbfile))
                self.pragma(f'temp_store_directory={sql_quote_ident(tmpdir)}')
        except:
            self.log(1, "WARN: cannot set 'pragma temp_store_directory', set SQLITE_TMPDIR envar instead if needed")


    ### DB control:

    def get_conn(self):
        if self.conn is None:
            if self.readonly:
                self.conn = sqlite3.connect(f'file:{self.dbfile}?mode=ro', uri=True)
            else:
                self.conn = sqlite3.connect(self.dbfile)
                # use explicit transaction management via BEGIN/COMMIT:
                self.conn.isolation_level = None
        return self.conn

    def get_sqlite_cmd(self):
        # TO DO: this should look more intelligently to find it in the python install path
        # e.g., check 'sys.prefix/bin', 'sys.exec_prefix/bin' or do a 'which sqlite3';
        # if we use a conda environment we get it automatically.
        return 'sqlite3'

    def close(self):
        if self.conn is not None:
            self.conn.close()
            self.conn = None
            self.user_functions = set()

    READONLY_REGEX = re.compile(r'^\s*(select|pragma)\s', re.IGNORECASE)
    
    def is_readonly_statement(self, statement):
        """Return True if the SQL 'statement' does not update the database.
        """
        return self.READONLY_REGEX.search(statement) is not None

    def execute(self, *args, **kwargs):
        if self.readonly and not self.is_readonly_statement(args[0]):
            # do nothing and return empty cursor:
            return sqlite3.Cursor(self.get_conn())
        else:
            return self.get_conn().execute(*args, **kwargs)

    def executemany(self, *args, **kwargs):
        if self.readonly and not self.is_readonly_statement(args[0]):
            # do nothing and return empty cursor:
            return sqlite3.Cursor(self.get_conn())
        else:
            return self.get_conn().executemany(*args, **kwargs)

    def in_transaction(self):
        """Return True if we are currently in the scope of a transaction.
        """
        return self.get_conn().in_transaction
            
    def ensure_transaction(self, type='DEFERRED'):
        """Ensure a transaction of 'type' has been started.  If we are currently
        inside a transaction, this is a no-op, which means a change of 'type'
        can only happen after a COMMIT has ended the current transaction.
        """
        if not self.in_transaction():
            self.log(2, f'BEGIN {type} TRANSACTION')
            self.execute(f'BEGIN {type} TRANSACTION')
            
    def commit(self):
        """Commit all updates of the current transaction to the database.
        This should only be called after complete and internally consistent
        top-level operations.
        """
        self.log(2, 'COMMIT updates to database...')
        if self.vector_store is not None:
            self.vector_store.commit()
        self.get_conn().commit()

    def pragma(self, expression):
        """Evaluate a PRAGMA 'expression' and return the result (if any).
        """
        res = list(self.execute('PRAGMA ' + expression))
        if len(res) == 0:
            return None
        elif len(res) == 1 and len(res[0]) == 1:
            return res[0][0]
        else:
            return res


    ### DB functions:

    @staticmethod
    def register_user_function(name, num_params, func, deterministic=False):
        """Old-style user function registration API, supported for backwards compatibility.
        """
        from kgtk.kypher.functions import SqlFunction
        SqlFunction(name, code=func, num_params=num_params, deterministic=deterministic).define()

    def load_user_function(self, name, num_params, func, deterministic=False):
        """Load the specified function 'name' into the current connection.
        Follows the sqlite3 API (which see).
        """
        conn = self.get_conn()
        try:
            if isinstance(func, type) and hasattr(func, 'register'):
                # we have a table-valued function described by a class object:
                func.register(conn)
            elif not deterministic:
                # try to avoid an error by not using the deterministic flag:
                conn.create_function(name, num_params, func)
            else:
                conn.create_function(name, num_params, func, deterministic=True)
        except sqlite3.NotSupportedError:
            # older SQLite, define it without 'deterministic':
            self.store.get_conn().create_function(name, num_params, func)
        # remember the function names we've loaded so far:
        self.user_functions.add(name)

    def load_aggregate_function(self, name, num_params, aggregate_class):
        """Load the specified aggregate function 'name' into the current connection.
        Follows the sqlite3 API (which see).
        """
        conn = self.get_conn()
        conn.create_aggregate(name, num_params, aggregate_class)
        self.user_functions.add(name)

    def get_user_functions(self):
        """Return the names of all user functions loaded into the current connection.
        """
        return self.user_functions


    ### DB properties and operations:
    
    def get_db_size(self):
        """Return the size of all currently allocated data pages in bytes.  This maybe smaller than
        the size of the database file if there were deletions that put pages back on the free list.
        Free pages can be reclaimed by running 'VACUUM', but that might require a substantial amount
        of available disk space if the current DB file is large.
        """
        return (self.pragma('page_count') - self.pragma('freelist_count')) * self.pragma('page_size')

    def get_db_free_size(self):
        """Return the size of all currently allocated but free data pages in bytes.
        """
        return self.pragma('freelist_count') * self.pragma('page_size')

    def has_table(self, table_name):
        """Return True if a table with name 'table_name' exists in the store.
        """
        schema = self.MASTER_TABLE
        columns = schema.columns
        query = """SELECT COUNT(*) FROM %s WHERE %s=?""" % (schema._name_, columns.name._name_)
        (cnt,) = self.execute(query, (table_name,)).fetchone()
        return cnt > 0

    def get_table_header(self, table_name):
        """Return the column names of 'table_name' as a list.  For graph tables, this list will be
        isomorphic to the parsed header line of the corresponding KGTK file.
        """
        result = self.execute('SELECT * FROM %s LIMIT 0' % table_name)
        return [col[0] for col in result.description]

    def get_table_row_count(self, table_name):
        for (cnt,) in self.execute('SELECT MAX(rowid) FROM %s' % table_name):
            return cnt or 0  # need to handle case of empty table without any rowids
        return 0

    def has_table_column(self, table_name, column_name):
        """Return True if table 'table_name' exists and has a column 'column_name'.
        """
        return self.has_table(table_name) and column_name in self.get_table_header(table_name)

    def get_table_definition_sql(self, table_name):
        """Return the 'CREATE TABLE <table_name> ...' statement used to create this table.
        """
        master = self.MASTER_TABLE._name_
        mcols = self.MASTER_TABLE.columns
        query = f"""SELECT {mcols.sql._name_} FROM {master} 
                    WHERE {mcols.type._name_}='table' AND {mcols.name._name_}=?"""
        for (stmt,) in self.execute(query, (table_name,)):
            return stmt
        return None

    def get_table_indexes_sql(self, table_name):
        """Return the 'CREATE INDEX ... ON <table_name> ...' statements used to index this table.
        """
        master = self.MASTER_TABLE._name_
        mcols = self.MASTER_TABLE.columns
        query = f"""SELECT {mcols.sql._name_} FROM {master} 
                    WHERE {mcols.type._name_}='index' AND {mcols.tbl_name._name_}=?"""
        return [stmt for (stmt,) in self.execute(query, (table_name,))]

    def sort_table(self, table_name, columns):
        """Sort the rows in table 'table_name' along one or more 'columns'.  Each column can be a simple
        column name or a full spec such as 'cast(node2 as int) desc' containing optional type coercion
        and sort directions (in which case the column name should be properly quoted if necessary).  This 
        creates a copy of 'table_name' into which the sorted rows will be inserted.  All indexes existing
        on the original table will also be recreated.  After sorting the freed up space from the original
        table and indexes is NOT automatically reclaimed.  To use it or free it either add additional
        data to the database or run the VACUUM command.  The value of sorting a table is primarily to
        create better data locality on disk, the sorted table will otherwise be identical to the original.
        """
        # we look up the actual definition statement to get all the types and column features correct:
        table_def = self.get_table_definition_sql(table_name)
        if table_def is None:
            raise KGTKException(f"table '{table_name}' does not exist")
        index_defs = self.get_table_indexes_sql(table_name)
        table_columns = self.get_table_header(table_name)
        sorted_table = f'{table_name}_sortEd_'
        sorted_def = re.sub(r"""(?i)^\s*(CREATE\s+TABLE\s+)['"`]?""" + re.escape(table_name) + r"""['"`]?(\s+.*)""",
                            f'\\1{sql_quote_ident(sorted_table)}\\2', table_def)
        if sorted_def == table_def:
            raise KGTKException(f"failed to sort '{table_name}', unexpected definition statement")
        sort_columns = [sql_quote_ident(col) if not re.search("""["` ]""", col) else col for col in listify(columns)]
        sort_stmt = f"""INSERT INTO {sql_quote_ident(sorted_table)} 
                        SELECT {', '.join(map(sql_quote_ident, table_columns))}
                        FROM {sql_quote_ident(table_name)}
                        ORDER BY {', '.join(sort_columns)}"""
        self.log(1, f"SORT table '{table_name}'...")
        self.execute(sorted_def)
        self.execute(sort_stmt)
        # ensure these counts are the same:
        if self.get_table_row_count(table_name) != self.get_table_row_count(sorted_table):
            raise KGTKException(f"something went wrong during sorting of '{table_name}' into '{sorted_table}'")
        self.log(1, f"DROP unsorted table '{table_name}'...")
        self.execute(f'DROP table {sql_quote_ident(table_name)}')
        # this can take some time on large tables:
        self.execute(f'ALTER TABLE {sql_quote_ident(sorted_table)} RENAME TO {sql_quote_ident(table_name)}')
        for index_def in index_defs:
            self.log(1, f"RE-CREATE index '{index_def}'...")
            self.execute(index_def)


    ### Schema manipulation:
    
    def kgtk_header_to_graph_table_schema(self, table_name, header):
        columns = sdict()
        for col in header:
            columns[col] = sdict['type': 'TEXT', '_name_': col]
        return sdict['_name_': table_name, 'columns': columns]

    def get_key_column(self, table_schema, error=True):
        """Return the name of the first column in 'schema' designated as a 'key',
        or raise an error if no key column has been designated (unless 'error' is False).
        This should only be used for single-key tables such as info tables.
        """
        for col in table_schema.columns.values():
            if col.get('key') == True:
                return col._name_
        if error:
            raise KGTKException('no key column defined')
        return None

    def get_table_definition(self, table_schema):
        """Generate an SQLite table definition for 'table_schema'.  Requires each column
        to have at least a 'type' property.  Optional 'default' properties will be translated
        into appropriate 'DEFAULT <value>' column constraints.  One or more columns with a
        'key' property will be translated into a 'PRIMARY KEY (col...)' constraint.  If there
        is more than one column with a key, they will be sorted by their values to order them.
        A 'without_rowid' property on the table will produce a 'WITHOUT ROWID' table (which
        requires a primary key to be legal!).  For some simple attribute tables such as 'labels',
        etc. that only index on 'node1' those might be more space efficient than regular tables.
        """
        colspecs = []
        keys = []
        for col in table_schema.columns.values():
            spec = sql_quote_ident(col._name_) + ' ' + col.type
            dflt = col.get('default')
            if dflt is not None:
                dflt = isinstance(dflt, (int, float)) and '{:+g}'.format(dflt) or '"%s"' % dflt
                spec += ' DEFAULT ' + dflt
            key = col.get('key')
            key and keys.append((col._name_, key))
            colspecs.append(spec)
        if len(keys) > 0:
            keys.sort(key=lambda x: x[1])
            keys = 'PRIMARY KEY (%s)' % ', '.join(map(lambda x: x[0], keys))
            colspecs.append(keys)
        temp = table_schema.get('temporary') and ' TEMPORARY' or ''
        norowid = table_schema.get('without_rowid') and ' WITHOUT ROWID' or ''
        return f'CREATE{temp} TABLE {table_schema._name_} ({", ".join(colspecs)}){norowid}'

    def get_table_index(self, table_or_schema, columns, unique=False):
        """Return a TableIndex object for an index on 'columns' for 'table_or_schema'.
        Create a unique or primary key index if 'unique' is True.
        """
        columns = [columns] if isinstance(columns, str) else columns # coerce to list
        index_spec = f'index: {", ".join([sql_quote_ident(col) for col in columns])}'
        if unique:
            index_spec += '//unique'
        return ispec.TableIndex(table_or_schema, index_spec)

    def get_column_list(self, *columns):
        return ', '.join([sql_quote_ident(col._name_) for col in columns])

    def get_full_column_list(self, table_schema):
        return ', '.join([sql_quote_ident(col._name_) for col in table_schema.columns.values()])


    ### File information and access:

    # Each fileinfo record is identified by a name key which defaults to the full
    # dereferenced realpath of the file from which the graph data was loaded.
    # If an alias was provided that name will be stored as the key instead.

    def normalize_file_path(self, file):
        # for stdin we key in on full filenames for now to also support 'stdin' as an alias:
        if file in ('-', '/dev/stdin'):
            return '/dev/stdin'
        else:
            return os.path.realpath(file)

    def is_standard_input(self, file):
        return self.normalize_file_path(file) == '/dev/stdin'

    def is_input_alias_name(self, name):
        """Return true if 'name' is a legal input alias.  We require aliases to not
        contain any path separators to distinguish them from file names which are
        stored as absolute pathnames in the file info table.
        """
        return name.find(os.sep) < 0
    
    def is_input_file_name(self, name):
        return not self.is_input_alias_name(name)

    def get_file_info(self, file, alias=None, exact=False):
        """Return a dict info structure for the file info for 'file' (or 'alias') or None
        if this file does not exist in the file table.  All column keys will be set in
        the result although some values may be None.  If 'exact', use 'file' as is and
        do not try to normalize it to an absolute path.  Matches based on 'file' will
        have preference over matches based on 'alias', for example, a file named 'graph'
        will match the entry for '/data/graph' (if that is its full name) before it
        matches an entry named by the alias 'mygraph', for example.
        """
        info = self.infotable.get_file_info(file)
        if info is None and not exact:
            file = self.normalize_file_path(file)
            info = self.infotable.get_file_info(file)
        if info is None and alias is not None:
            info = self.infotable.get_file_info(alias)
        return info

    def get_normalized_file(self, file, alias=None, exact=False):
        """Return the stored normalized name of 'file' (or 'alias') or None
        if this file does not exist in the file table.
        """
        info = self.get_file_info(file, alias=alias, exact=exact)
        return info and info.file or None
        
    def set_file_info(self, _file, **kwargs):
        # TRICKY: we use '_file' so we can also use and update 'file' in 'kwargs'
        self.infotable.set_file_info(_file, kwargs)

    def update_file_info(self, _file, **kwargs):
        self.infotable.update_file_info(_file, kwargs)
        
    def drop_file_info(self, file):
        """Delete the file info record for 'file'.
        IMPORTANT: this does not delete any graph data associated with 'file'.
        """
        self.infotable.drop_file_info(file)
        
    def describe_file_info(self, file, out=sys.stderr):
        """Describe a single 'file' (or its info) to 'out'.
        """
        info = isinstance(file, dict) and file or self.get_file_info(file)
        out.write('%s:\n' % info.file)
        out.write('  size:  %s' % (info.size and format_memory_size(int(info.size)) or '???   '))
        out.write('   \tmodified:  %s' % time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(float(info.modtime))))
        out.write('   \tgraph:  %s\n' % info.graph)
        if info.comment:
            out.write('  comment:  %s\n' % info.comment)

    def describe_file_info_table(self, out=sys.stderr):
        """Describe all files in the INFO_TABLE to 'out'.
        """
        for info in self.infotable.get_all_file_infos():
            self.describe_file_info(info, out=out)

    def set_file_alias(self, file, alias):
        """Set the file column of the file info identified by 'file' (or 'alias') to 'alias'.
        Raises an error if no relevant file info could be found, or if 'alias' is already
        used in a different file info (in which case it wouldn't be a unique key anymore).
        """
        finfo = self.get_file_info(file, alias=alias)
        if finfo is None:
            raise KGTKException('cannot set alias for non-existent file: %s' % file)
        ainfo = self.get_file_info(alias, exact=True)
        if ainfo is not None and ainfo != finfo:
            # this can happen if we imported 'file' without an alias, then another file
            # with 'alias', and then we try to associate 'alias' to 'file':
            raise KGTKException('alias %s is already in use for different file' % alias)
        # update current file name to 'alias':
        self.update_file_info(finfo.file, file=alias)

    def set_file_comment(self, file, comment):
        """Set the comment property for 'file'.
        """
        # We might need some text normalization here:
        self.update_file_info(file, comment=comment)

    def get_file_graph(self, file):
        """Return the graph table name created from the data of 'file'.
        """
        return self.get_file_info(file).graph

    def get_graph_files(self, table_name):
        """Return the list of all files whose data is represented by 'table_name'.
        Generally, there will only be one, but it is possible that different versions
        of a file (e.g., compressed vs. uncompressed) created the same underlying data
        which we could detect by running a sha hash command on the resulting tables.
        """
        files = self.infotable.get_objects(table_name, InfoTable.graph)
        return [self.infotable.get_value(f, InfoTable.file) for f in files]


    ### Graph information and access:

    # TO DO: add 'bump_timestamp' so we can easily track when this graph was last used
    #        add 'update_xxx_info' methods that only change not None fields
    
    def get_graph_info(self, table_name):
        """Return a dict info structure for the graph stored in 'table_name' (there can only be one),
        or None if this graph does not exist in the graph table.  All column keys will be set
        although some values may be None.
        """
        return self.infotable.get_graph_info(table_name)

    def set_graph_info(self, table_name, **kwargs):
        self.infotable.set_graph_info(table_name, kwargs)
    
    def update_graph_info(self, table_name, **kwargs):
        self.infotable.update_graph_info(table_name, kwargs)
        
    def drop_graph_info(self, table_name):
        """Delete the graph info record for 'table_name'.
        IMPORTANT: this does not delete any graph data stored in 'table_name'.
        """
        self.infotable.drop_graph_info(table_name)

    def describe_graph_info(self, graph, out=sys.stderr):
        """Describe a single 'graph' (or its info) to 'out'.
        """
        info = isinstance(graph, dict) and graph or self.get_graph_info(graph)
        out.write('%s:\n' % info.name)
        out.write('  size:  %s' % format_memory_size(int(info.size)))
        out.write('   \tcreated:  %s\n' % time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(float(info.acctime))))
        out.write('  header:  %s\n' % info.header)

    def describe_graph_info_table(self, out=sys.stderr):
        """Describe all graphs in the INFO_TABLE to 'out'.
        """
        for info in self.infotable.get_all_graph_infos():
            self.describe_graph_info(info, out=out)

    def get_graph_table_schema(self, table_name):
        """Get a graph table schema definition for graph 'table_name'.
        """
        info = self.get_graph_info(table_name)
        header = eval(info.header)
        return self.kgtk_header_to_graph_table_schema(table_name, header)

    def get_graph_indexes(self, table_name):
        """Return the list of indexes currently defined for graph 'table_name'.
        """
        info = self.get_graph_info(table_name)
        if info is None:
            # see if 'table_name' is a defined or loaded virtual table function:
            if SqlFunction.is_virtual_graph(table_name) or table_name in self.get_user_functions():
                return []
            raise KGTKException(f"INTERNAL ERROR: missing graph info for '{table_name}'")
        indexes = [ispec.TableIndex.decode(idx) for idx in listify(info.index)]
        return indexes

    def has_graph_index(self, table_name, index):
        """Return True if graph 'table_name' has an index that subsumes 'index'.
        """
        for idx in self.get_graph_indexes(table_name):
            if idx.subsumes(index) and not index.redefines(idx):
                return True
        else:
            return False

    def get_vector_indexes(self, table_name):
        """Return the list of vector indexes currently defined for graph 'table_name'.
        """
        return [idx for idx in self.get_graph_indexes(table_name) if isinstance(idx, ispec.VectorIndex)]

    def is_vector_column(self, table_name, column):
        """Return True if 'column' in 'table_name' is a vector column.
        """
        for vindex in self.get_vector_indexes(table_name):
            if column in vindex.index.columns:
                return True
        else:
            return False

    def ensure_graph_index(self, table_name, index, explain=False, commit=True):
        """Ensure a qualifying 'index' for 'table_name' already exists or gets created.
        Checks whether the existing index is at least as selective as requested, for
        example, an existing index on columns (node1, node2) will qualify even if 'index'
        has 'node1' as its only column.
        """
        if not self.has_graph_index(table_name, index):
            if self.readonly:
                return
            self.ensure_transaction()
            loglevel = 0 if explain else 1
            indexes = self.get_graph_indexes(table_name)
            # delete anything that is redefined by this 'index':
            for idx in indexes[:]:
                if index.redefines(idx) and not explain:
                    self.drop_graph_index(table_name, idx)
            indexes = self.get_graph_indexes(table_name)
            # we also measure the increase in allocated disk space here:
            oldsize = self.get_db_size()
            index.create_index(self, explain=explain)
            idxsize = self.get_db_size() - oldsize
            ginfo = self.get_graph_info(table_name)
            ginfo.size = int(ginfo.size) + idxsize
            if not explain:
                # we may have added/deleted some, so recompute the list:
                indexes = self.get_graph_indexes(table_name) + [index]
                indexes = [ispec.TableIndex.encode(idx) for idx in indexes]
                self.update_graph_info(table_name, index=indexes)
                self.update_graph_info(table_name, size=ginfo.size)
                if commit:
                    # top-level OP, commit DB and info updates:
                    self.commit()

    def ensure_graph_index_for_columns(self, table_name, columns, unique=False, explain=False, commit=True):
        """Ensure an index for 'table_name' on 'columns' already exists or gets created.
        Checks whether the existing index is at least as selective as requested, for example,
        an existing index on columns (node1, node2) will qualify even if only node1 is requested.
        """
        index = self.get_table_index(table_name, columns, unique=unique)
        self.ensure_graph_index(table_name, index, explain=explain, commit=commit)

    def number_of_graphs(self):
        """Return the number of graphs currently stored in 'self'.
        """
        return len(self.infotable.get_objects('graph', InfoTable.type))

    def new_graph_table(self):
        """Return a new table name to be used for representing a graph.
        """
        graphid = (self.number_of_graphs() + 1)
        # search for an open ID (we might have gaps due to deletions):
        while True:
            table = 'graph_%d' % graphid
            if not self.has_table(table):
                return table
            graphid += 1

    def determine_graph_action(self, file, alias=None, error=True):
        """Determine which action to perform for the KGTK graph indicated by input 'file' (or 'alias').
        Returns one of 'add', 'replace', 'reuse' or 'error'.  Raises an exception for error cases in
        case 'error' was set to True (the default).

        Returns 'add' if no matching file info based on 'file/alias' could be found, in which case
        the data needs to be newly imported.

        Returns 'reuse' if a matching file info was found and 'file' is an existing regular file whose
        properties match exactly what was previously loaded, or is not an existing regular file in which
        case its properties cannot be checked.  This latter case allows us to delete large input files
        after import without losing the ability to query them, or to query files by using their alias
        instead of a real filename.

        Returns 'replace' if a matching file info was found and 'file' is an existing regular file
        whose properties do not match what was previously loaded, or if 'file' names standard input.
        If so an obsolete graph table for 'file' will have to be removed before new data gets imported.

        Checks for errors such as invalid alias names, aliases that are already in use for other
        inputs, and cases where an existing file might conflict with an existing input alias.
        """
        if alias is not None and not self.is_input_alias_name(alias):
            if error:
                raise KGTKException(f'invalid input alias name: {alias}')
            else:
                return 'error'
        
        info = self.get_file_info(file, alias=alias)
        if info is None:
            return 'add'
        
        is_aliased = self.is_input_alias_name(info.file)
        defines_alias = alias is not None
        if defines_alias:
            alias_info = self.get_file_info(alias, exact=True)
            if alias_info is not None and info.file != alias_info.file:
                if error:
                    raise KGTKException(f"input alias '{alias}' already in use")
                else:
                    return 'error'
        
        if self.is_standard_input(file):
            # we never reuse plain stdin, it needs to be aliased to a new name for that:
            return 'replace'
        
        if os.path.exists(file):
            if is_aliased and not defines_alias:
                if error:
                    raise KGTKException(f"input '{file}' conflicts with existing alias; "+
                                        f"to replace use explicit '--as {info.file}'")
                else:
                    return 'error'
            if int(info.size) !=  os.path.getsize(file):
                return 'replace'
            if float(info.modtime) != os.path.getmtime(file):
                return 'replace'
            # don't check md5sum for now
        return 'reuse'

    def has_graph(self, file, alias=None):
        """Return True if the KGTK graph represented/named by 'file' (or its 'alias' if not None)
        has already been imported and is up-to-date (see 'determine_graph_action' for the full story).
        """
        return self.determine_graph_action(file, alias=alias, error=False) == 'reuse'

    def add_graph(self, file, alias=None, index_specs=None, append=None, commit=True):
        """Import a graph from 'file' (and optionally named by 'alias') unless a matching
        graph has already been imported earlier according to 'has_graph' (which see).
        'index_specs' is a list of index specifications relevant to the import of this graph.
        'append' can be one or more files whose data should be appended to the graph identified
        by 'file'.  If any to-be-appended file already exists in the info table, it will simply
        be ignored and the data reused.  Updating of appended data is not supported, however,
        the whole graph can be replaced with data from a new 'file'.  Appended data needs
        to match the columns of the data that it is appended to, however, no other checking
        for duplicates, etc. is performed.  Data is simply treated as if it had been appended
        to the original data file.  Reordering of columns is supported for the cost of some
        import speed.
        """
        # TO DO: make this less messy and hairy...
        graph_action = self.determine_graph_action(file, alias=alias)
        file_info = self.get_file_info(file, alias=alias)

        # determine what we have to do with the supplied file if anything:
        if graph_action == 'reuse':
            if alias is not None:
                # this allows us to do multiple renamings (no-op if we are readonly):
                self.set_file_alias(file, alias)
        if self.readonly and graph_action != 'reuse':
            raise KGTKException(f"cannot import or update {file} in read-only mode")

        # determine the graph table we are operating on:
        if graph_action != 'reuse':
            table = self.new_graph_table()
        else:
            table = file_info.graph

        # vector tables are handled in two steps, vector data import and indexing, and we have
        # to figure out whether one or both steps need to be redone upon index redefinition:
        vector_spec = self.get_vector_index_specs(table, index_specs)
        vector_spec = self.normalize_vector_index_specs(table, vector_spec) if vector_spec else None
        if graph_action == 'reuse' and vector_spec and not self.readonly:
            for idx in self.get_vector_indexes(table):
                if vector_spec.redefines_store(idx):
                    graph_action = 'replace'
                    table = self.new_graph_table()
                    vector_spec.table = table
                    break

        if graph_action != 'reuse':
            # import main data:
            self.add_graph_data(table, file, alias=alias, index_specs=index_specs, append=False)
        if graph_action == 'replace':
            # delete any old graph data *after* we replaced with new version to not lose anything in case of error:
            # TRICKY: we already pointed the new file_info.graph to the new table, so it won't be deleted here:
            self.drop_graph(file_info.graph)

        if vector_spec:
            # handle step 2 of vector index creation:
            self.ensure_graph_index(table, vector_spec, commit=False)
        if commit and self.in_transaction():
            # top-level OP, commit successful DB, index and info updates:
            if vector_spec:
                self.get_vector_store().commit()
            self.commit()

        # append any additional data files:
        for afile in listify(append):
            graph_action = self.determine_graph_action(afile)
            file_info = self.get_file_info(afile)
            if graph_action == 'replace':
                if self.readonly:
                    graph_action = 'reuse'
                else:
                    raise KGTKException(f"cannot replace {afile} in append mode")
            if graph_action == 'reuse' and file_info.graph != table:
                # each file must correspond to exactly one graph:
                raise KGTKException(f"{afile} is already in use by a different graph: {table}")
            if graph_action == 'add':
                self.add_graph_data(table, afile, index_specs=index_specs, append=True)
                if commit:
                    # top-level OP, commit successful DB and info updates:
                    self.commit()

    def add_graph_data(self, table, file, alias=None, index_specs=None, append=False):
        """Low-level implementation of 'add_graph'.  Import data for graph 'table' from 'file'
        or add it to an existing graph if 'append' is True.  Uses a fast direct import
        if possible, otherwise falls back on a 2x slower csv.reader-based import of the data.
        'index_specs' is a list of index specifications relevant to the import of this graph.
        """
        file = self.normalize_file_path(file)
        oldsize = self.get_db_size()
        vector_specs = self.get_vector_index_specs(table, index_specs)
        if len(vector_specs) > 0:
            if append:
                raise KGTKException('cannot append to vector data')
            self.ensure_transaction()
            # perform step 1 of vector data import, step 2 indexing is done by the caller:
            self.import_graph_vector_data_via_csv(table, file, index_specs=vector_specs)
        else:
            fast_import = False
            # fast import runs in a separate process at the shell level, so we cannot use it here
            # if we are currently inside a transaction on this connection, since the DB will be locked:
            # TO DO: see if we can perform explicit post-hoc cleanup in case something crashes here,
            #        given that our local transactions will not handle this case:
            if not self.in_transaction():
                try:
                    # try fast shell-based import first, but if that is not applicable or supported...
                    fast_import = self.import_graph_data_via_import(table, file, append=append)
                except (KGTKException, sh.CommandNotFound):
                    pass
            self.ensure_transaction()
            if not fast_import:
                # ...fall back on CSV-based import which is more flexible but about 2x slower:
                self.import_graph_data_via_csv(table, file, append=append)
        graphsize = self.get_db_size() - oldsize
        # this isn't really needed, but we store it for now:
        header = str(self.get_table_header(table))
        if self.is_standard_input(file):
            self.set_file_info(file, size=0, modtime=time.time(), graph=table)
        else:
            self.set_file_info(file, size=os.path.getsize(file), modtime=os.path.getmtime(file), graph=table)
        self.set_graph_info(table, header=header, size=graphsize, acctime=time.time())
        if alias is not None:
            self.set_file_alias(file, alias)
            
    def drop_graph(self, table_name):
        """Delete the graph 'table_name' and all its associated info records.
        """
        if self.readonly:
            return
        self.log(1, f'DROP graph data table {table_name}')
        # delete any remaining supporting file infos pointing to 'table_name':
        for file in self.get_graph_files(table_name):
            self.drop_file_info(file)
        # delete the graph info:
        self.drop_graph_info(table_name)
        # now delete the graph table and all associated indexes:
        if self.has_table(table_name):
            self.execute('DROP TABLE %s' % table_name)

    def drop_graph_index(self, table_name, index):
        """Delete 'index' for graph 'table_name' and its associated info records.
        """
        if self.readonly:
            return
        ginfo = self.get_graph_info(table_name)
        indexes = self.get_graph_indexes(table_name)
        if index not in indexes:
            raise KGTKException(f'No such index for {table_name}: {index}]')
        oldsize = self.get_db_size()
        index.drop_index(self, explain=False)
        idxsize = oldsize - self.get_db_size()
        # we might have updated them, so recompute the list:
        indexes = self.get_graph_indexes(table_name)
        indexes.remove(index)
        indexes = [ispec.TableIndex.encode(idx) for idx in indexes]
        ginfo.size = int(ginfo.size) - idxsize
        self.update_graph_info(table_name, index=indexes, size=ginfo.size)

    def drop_graph_indexes(self, table_name, index_type=None, commit=True):
        """Delete all indexes for graph 'table_name'.  If 'index_type' is not None,
        restrict to indexes of that type (can be a short name or a class).
        """
        self.ensure_transaction()
        if isinstance(index_type, str):
            index_type = ispec.TableIndex.get_index_type_class(index_type)
        for index in self.get_graph_indexes(table_name)[:]:
            if index_type is None or isinstance(index, index_type):
                self.drop_graph_index(table_name, index)
        if commit:
            # top-level OP, commit DB and info updates:
            self.commit()


    ### Data import:
    
    def import_graph_data_via_csv(self, table, file, append=False):
        """Import 'file' into 'table' using Python's csv.reader.  This is safe and properly
        handles conversion of different kinds of line endings, but 2x slower than direct import.
        """
        if self.readonly:
            return
        self.log(1, 'IMPORT graph via csv.reader into table %s from %s ...' % (table, file))
        if self.is_standard_input(file):
            file = sys.stdin
        with open_to_read(file) as inp:
            csvreader = csv.reader(inp, dialect=None, delimiter='\t', quoting=csv.QUOTE_NONE)
            header = next(csvreader)
            schema = self.kgtk_header_to_graph_table_schema(table, header)
            column_list = ''
            if append:
                if self.has_table(table):
                    table_header = self.get_table_header(table)
                    if set(table_header) != set(header):
                        raise KGTKException('can only append data with the same set of columns')
                    if table_header != header:
                        # new data columns are in a different order, so we need an explicit column list:
                        column_list = f' ({self.get_full_column_list(self.kgtk_header_to_graph_table_schema(table, header))})'
                else:
                    append = False
            if not append:
                self.execute(self.get_table_definition(schema))
            insert = f'INSERT INTO {table}{column_list} VALUES ({",".join(["?"] * len(header))})'
            self.executemany(insert, csvreader)

    def import_graph_data_via_import(self, table, file, append=False):
        """Use the sqlite shell and its import command to import 'file' into 'table'.
        This will be about 2+ times faster and can exploit parallelism for decompression.
        This is only supported for Un*x for now and requires a named 'file'.
        Return True if the import finished successfully, raise exception otherwise.
        """
        if self.readonly:
            return True
        if os.name != 'posix':
            raise KGTKException("not yet implemented for this OS: '%s'" % os.name)
        # generalizing this to work for stdin would be possible, but it would significantly complicate
        # matters, since we also have to check for multi-char line endings at which point we can't
        # simply abort to 'import_graph_data_via_csv' but would have to buffer and resupply the read data:
        if not isinstance(file, str) or not os.path.exists(file) or self.is_standard_input(file):
            raise KGTKException('only implemented for existing, named files')
        # make sure we have the Unix commands we need:
        catcmd = get_cat_command(file, _piped=True)
        tail = sh.Command('tail')
        sqlite3 = sh.Command(self.get_sqlite_cmd())
        isplain = os.path.basename(catcmd._path) == b'cat'
        
        # This is slightly more messy than we'd like it to be: sqlite can create a table definition
        # for a non-existing table from the header row, but it doesn't seem to handle just any weird
        # column name we give it there, so we read the header and create the table ourselves;
        # however, sqlite doesn't have an option to then skip the header, so we need to use 'tail';
        # also, eventually we might want to supply more elaborate table defs such as 'without rowid';
        # finally, we have to guard against multi-character line-endings which can't be handled right:
        with open_to_read(file, 'rt') as inp:
            #csvreader = csv.reader(inp, dialect=None, delimiter='\t', quoting=csv.QUOTE_NONE)
            header = inp.readline()
            if inp.newlines != '\n':
                # SQLite import can only handle single-character line endings, if we import anyway,
                # \r winds up in the values of the last column; we also can't handle \r by itself
                # (which should be rare - not used since MacOS X), since that will not work with 'tail'.
                # We could handle both cases by mapping to \n with 'tr', but that introduces an extra
                # pipe and command complication - maybe later:
                raise KGTKException('unsupported line endings')
            header = header[:-1].split('\t')
            schema = self.kgtk_header_to_graph_table_schema(table, header)
            if append:
                if self.has_table(table):
                    if self.get_table_header(table) != header:
                        raise KGTKException('can only fast-append data with identical columns')
                else:
                    append = False
            if not append:
                # create the table in a separate process and not on the local connection
                # to not create any locking issues with local transactions:
                sqlite3(self.dbfile, self.get_table_definition(schema))
        
        separators = '\\t \\n'
        args = ['-cmd', '.mode ascii', '-cmd', '.separator ' + separators,
                self.dbfile, '.import /dev/stdin %s' % table]

        self.log(1, 'IMPORT graph directly into table %s from %s ...' % (table, file))
        try:
            if isplain:
                tailproc = tail('-n', '+2', file, _piped=True)
            else:
                tailproc = tail(catcmd(), '-n', '+2', _piped=True)
            # we run this asynchronously, so we can kill it in the cleanup clause:
            sqlproc = sqlite3(tailproc, *args, _bg=True)
            sqlproc.wait()
        finally:
            # make sure we kill this process in case we had a user interrupt, however,
            # getting this condition right so we don't hang and don't break was tricky,
            # since there is various machinery under the hood which leads to additional
            # waiting (we can't call is_alive or access sqlproc.exit_code):
            if sqlproc is not None and sqlproc.process.exit_code is None:
                sqlproc.terminate()
        return True

                
    # we need to be able to comfortably fit that many un/parsed vectors into RAM:
    VECTOR_IMPORT_CHUNKSIZE = 100000

    def get_vector_store(self):
        if self.vector_store is None:
            import kgtk.kypher.vecstore as vs
            self.vector_store = vs.MasterVectorStore(self)
        return self.vector_store

    def get_vector_index_specs(self, graph, index_specs):
        vector_specs = []
        for index_spec in listify(index_specs):
            index_spec = ispec.get_normalized_index_mode(index_spec)
            if isinstance(index_spec, list):
                for spec in index_spec:
                    spec = ispec.TableIndex(graph, spec)
                    if isinstance(spec, ispec.VectorIndex):
                        vector_specs.append(spec)
        return vector_specs

    def normalize_vector_index_specs(self, graph, index_specs):
        """Normalize all 'index_specs' for 'graph' into a single one containing all the
        relevant information.  This is somewhat different than other index specs, since
        we assume there can be at most one index per vector column of each graph.
        """
        index_specs = listify(index_specs)
        if len(index_specs) == 0:
            # use this as the default, even though 'index_specs' shouldn't be empty:
            index_specs = [ispec.TableIndex(graph, 'vector: node2/fmt=auto')]
        vector_spec = index_specs[0]
        for ospec in index_specs[1:]:
            # merge and override from any subsequent specs:
            for col, spec in ospec.index.columns.items():
                vector_spec.index.columns[col] = spec
        return vector_spec

    def import_graph_vector_data_via_csv(self, table, file, index_specs=None):
        """Import 'file' into 'table' using Python's csv.reader.  Parses and stores any vector columns
        in binary format into the respective vector cache and blanks out the corresponding source field(s).
        """
        if self.readonly:
            return
        self.log(1, f'IMPORT graph vectors into table {table} and associated vector cache from {file}')
        if self.is_standard_input(file):
            file = sys.stdin

        vector_spec = self.normalize_vector_index_specs(table, index_specs)
        vector_columns = list(vector_spec.index.columns.keys())
        vstore = self.get_vector_store()
        
        with open_to_read(file) as inp:
            csvreader = csv.reader(inp, dialect=None, delimiter='\t', quoting=csv.QUOTE_NONE)
            header = next(csvreader)
            schema = self.kgtk_header_to_graph_table_schema(table, header)

            for vcol in vector_columns:
                if vcol not in header:
                    raise KGTKException(f"vector column '{vcol}' does not exist in {file}")
            for vcol in vector_columns:
                colvs = vstore.get_vector_store(table, vcol, index_spec=vector_spec)
                schema = colvs.get_graph_table_schema(schema)
                # TO DO: maybe rename and defer the drop until after import is successfully completed:
                colvs.drop_store_data()
            
            insert = None
            while True:
                chunk = list(itertools.islice(csvreader, self.VECTOR_IMPORT_CHUNKSIZE))
                for vcol in vector_columns:
                    colvs = vstore.get_vector_store(table, vcol)
                    chunk = colvs.import_vectors(header.index(vcol), chunk)
                    
                # if vector-adapted 'schema' is not None, add remaining/reformatted data fields to the graph cache:
                if schema is not None and insert is None:
                    insert = f'INSERT INTO {table} VALUES ({",".join(["?"] * len(schema.columns))})'
                    # defer this until we added some vectors to the cache in case anything goes wrong:
                    self.execute(self.get_table_definition(schema))
                if len(chunk) == 0:
                    break
                if schema is not None:
                    self.executemany(insert, chunk)
        # make sure we close any open files:
        vstore.close()

    def vector_column_to_sql(self, table, column, table_alias=None):
        """Return an SQL expression that can be used to generate the actual vectors
        corresponding to the vector-indexed 'column' variable of 'table'.
        """
        return self.get_vector_store().vector_column_to_sql(table, column, table_alias=table_alias)
    
                
    def shell(self, *commands):
        """Execute a sequence of sqlite3 shell 'commands' in a single invocation
        and return stdout and stderr as result strings.  These sqlite shell commands
        are not invokable from a connection object, they have to be entered via 'sh'.
        """
        # we can't easily guard this for readonly use, so for now we have
        # to rely on the fact that it is only used safely within this module:
        sqlite3 = sh.Command(self.get_sqlite_cmd())
        args = []
        for cmd in commands[0:-1]:
            args.append('-cmd')
            args.append(cmd)
        args.append(self.dbfile)
        args.append(commands[-1])
        proc = sqlite3(*args)
        return proc.stdout, proc.stderr

    def explain(self, sql_query, parameters=None, mode='plan'):
        """Generate a query execution plan explanation for 'sql_query' and return it as a string.
        If the query contains any parameter place holders, a set of actual 'parameters' needs to
        be supplied.  'mode' needs to be one of 'plan' (the default), 'full' or 'expert'.  Except
        for 'plan' mode, 'sql_query' may not contain any KGTK user function references.
        """
        if mode == 'plan':
            plan = self.get_query_plan(sql_query, parameters)
            return self.get_query_plan_description(plan)
        # for the other two modes we use an SQLite shell command which doesn't require parameters:
        elif mode == 'full':
            out, err = self.shell('EXPLAIN ' + sql_query)
        elif mode == 'expert':
            out, err = self.shell('.expert', sql_query)
        else:
            raise KGTKException('illegal explanation mode: %s' % str(mode))
        return out.decode('utf8')

    def get_query_plan(self, sql_query, parameters=None):
        """Return a list of query plan steps for 'sql_query' and 'parameters'.
        Each step is a tuple of id, parent_id and description string.
        """
        explain_cmd = 'EXPLAIN QUERY PLAN ' + sql_query
        parameters = parameters is not None and parameters or ()
        plan = []
        for node, parent, aux, desc in self.execute(explain_cmd, parameters):
            plan.append((node, parent, desc))
        return plan

    def get_query_plan_description(self, plan):
        """Return a textual description of a query 'plan' generated by 'get_query_plan'.
        This closely mirrors the top-level rendering of SQLite, but not exactly so.
        """
        node_depths = {}
        out = io.StringIO()
        out.write('QUERY PLAN\n')
        for node, parent, desc in plan:
            depth = node_depths.get(node)
            if depth is None:
                depth = node_depths.get(parent, 0) + 1
                node_depths[node] = depth
            out.write('|  ' * (depth-1))
            out.write('|--')
            out.write(desc)
            out.write('\n')
        return out.getvalue()

    def suggest_indexes(self, sql_query):
        explanation = self.explain(sql_query, mode='expert')
        indexes = []
        index_regex = re.compile(r'\s*CREATE\s+INDEX\s+(?P<name>[^\s]+)'
                                 + r'\s+ON\s+(?P<table>[^\s(]+)'
                                 + r'\s*\(\s*(?P<columns>[^\s,)]+(\s*,\s*[^\s,)]+)*)\s*\)',
                                 re.IGNORECASE)
        split_regex = re.compile(r'\s*,\s*')
        for line in explanation.splitlines():
            m = index_regex.match(line)
            if m is not None:
                name = m['name']
                table = m['table']
                columns = m['columns']
                columns = split_regex.split(columns)
                indexes.append((name, table, columns))
        return indexes


class InfoTable(KgtkInfoTable):
    """Schema-free info table using a fine-grained KGTK edge data model.
    All node1/label/node2/id values are stored as strings, so any conversions
    of numbers, etc. need to be handled explicitly by the user of this table.
    """

    # we mostly use the old column headers as labels for simplicity:
    acctime = 'acctime'
    comment = 'comment'
    file = 'file'
    graph = 'graph'
    header = 'header'
    index = 'index'
    md5sum = 'md5sum'
    modtime = 'modtime'
    name = 'name'
    shasum = 'shasum'
    size = 'size'
    type = 'type'

    # info object templates:
    FILE_INFO = sdict([(key, None) for key in [type, file, size, modtime, graph, comment]])
    GRAPH_INFO = sdict([(key, None) for key in [type, name, header, size, acctime, index]])

    def init_table(self):
        """If the info table doesn't exist yet, define it from its schema.
        """
        if not self.store.has_table(self.table):
            self.store.ensure_transaction()
            super().init_table()
            self.transfer_fileinfo()
            self.transfer_graphinfo()
            self.store.commit()

    def make_file_id(self, path):
        """Create a file object ID uniquely based on a file's 'path'.
        """
        return self.make_object_id('file-', path)

    def transfer_fileinfo(self):
        """Transfer information from an old-style fileinfo table.
        """
        store = self.store
        import kgtk.kypher.infotable as oldit
        if store.has_table(oldit.FILE_TABLE._name_):
            fileinfo = oldit.InfoTable(store, oldit.FILE_TABLE)
            for info in fileinfo.get_all_infos():
                self.set_file_info(info.file, info)

    def transfer_graphinfo(self):
        """Transfer information from an old-style graphinfo table.
        """
        store = self.store
        import kgtk.kypher.infotable as oldit
        if store.has_table(oldit.GRAPH_TABLE._name_):
            graphinfo = oldit.InfoTable(store, oldit.GRAPH_TABLE)
            for info in graphinfo.get_all_infos():
                # translate index list into multi-valued edge:
                info[InfoTable.index] = [ispec.TableIndex.encode(idx) for idx in ispec.TableIndex.decode(info.indexes)]
                del info['indexes']
                self.set_graph_info(info.name, info)

    # TO DO: add LRU caching and maintenance
    def get_info(self, obj, info=None):
        """Retrieve all edges that start with 'obj' and create a dict from their
        'label/node2' value pairs.  Multi-valued labels will be converted into lists.
        If 'info' is provided, populate it instead of creating a new dict.
        """
        info = info if info is not None else sdict()
        for _, key, value, _ in self.get_edges(obj):
            if info.get(key) is not None:
                info[key] = listify(info[key]) + [value]
            else:
                info[key] = value
        return info if len(info) > 0 else None

    def set_info(self, obj, info):
        """Set edge values for 'obj' based on non-null key/value pairs in 'info'.
        List values will automatically be translated into multi-valued edges.
        This coerces all values to strings to avoid precision issues with floats.
        """
        for key, value in info.items():
            if value is None:
                continue
            if not isinstance(value, str) and hasattr(value, '__iter__'):
                # we have a multi-valued attribute:
                for i, val in enumerate(value):
                    if i == 0:
                        self.set_value(obj, key, str(val))
                    else:
                        self.add_value(obj, key, str(val))
            else:
                self.set_value(obj, key, str(value))
    
    def update_info(self, obj, info):
        """Just like 'set_info' but a no-op if 'obj' does not exist in the table.
        """
        for edge in self.get_edges(obj):
            self.set_info(obj, info)
            return
    
    def drop_info(self, obj):
        """Delete all edges that have 'obj' as their node1.
        """
        self.delete_edges(obj)
        

    def get_file_info(self, path):
        """Return a dict info structure for the file with 'path' (there can only be one),
        or None if this file does not exist in the info table.  All supported keys will
        be set although some values may be None.
        """
        file = self.get_object(path, InfoTable.file)
        return self.get_info(file, info=copy.deepcopy(self.FILE_INFO)) if file is not None else None

    def set_file_info(self, path, info):
        """Set all 'info' for the file with 'path'.  Create a new file ID if necessary.
        """
        file = self.get_object(path, InfoTable.file)
        if file is None:
            # new file, generate an ID for it:
            file = self.make_file_id(path)
            info[InfoTable.file] = path
            # also report the type:
            info[InfoTable.type] = 'file'
        self.set_info(file, info)
    
    def update_file_info(self, path, info):
        file = self.get_object(path, InfoTable.file)
        if file is not None:
            self.update_info(file, info)
    
    def drop_file_info(self, path):
        """Delete the graph info record for the file with 'path'.
        """
        file = self.get_object(path, InfoTable.file)
        if file is not None:
            self.drop_info(file)

            
    def get_graph_info(self, graph):
        """Return a dict info structure for 'graph' (there can only be one), or None
        if this graph does not exist in the info table.  All supported keys will be set
        although some values may be None.
        """
        return self.get_info(graph, info=copy.deepcopy(self.GRAPH_INFO))

    def set_graph_info(self, graph, info):
        info[InfoTable.type] = 'graph'
        # add a default 'name' attribute:
        info[InfoTable.name] = info.get(InfoTable.name, graph)
        self.set_info(graph, info)
    
    def update_graph_info(self, graph, info):
        self.update_info(graph, info)
        
    def drop_graph_info(self, graph):
        """Delete the graph info record for 'table_name'.
        """
        self.drop_info(graph)
        
    
    def get_all_infos(self, type=None):
        infos = []
        for obj in self.get_all_objects():
            info = None
            if type == 'file':
                info = copy.deepcopy(self.FILE_INFO)
            elif type == 'graph':
                info = copy.deepcopy(self.GRAPH_INFO)
            info = self.get_info(obj, info=info)
            if type == None or info.type == type:
                infos.append(info)
        return infos

    def get_all_file_infos(self):
        return self.get_all_infos(type='file')

    def get_all_graph_infos(self):
        return self.get_all_infos(type='graph')


### Experiments:

"""
>>> store = cq.SqliteStore('/data/tmp/store.db', create=True)
>>> store.add_graph('kgtk/tests/data/kypher/graph.tsv')

>>> cq.pp.pprint(list(store.execute('select * from graph_1')))
[   ('Hans', 'loves', 'Molly', 'e11'),
    ('Otto', 'loves', 'Susi', 'e12'),
    ('Joe', 'friend', 'Otto', 'e13'),
    ('Joe', 'loves', 'Joe', 'e14'),
    ('Hans', 'name', '"Hans"', 'e21'),
    ('Otto', 'name', '"Otto"', 'e22'),
    ('Joe', 'name', '"Joe"', 'e23'),
    ('Molly', 'name', '"Molly"', 'e24'),
    ('Susi', 'name', '"Susi"', 'e25')]

>>> cq.pp.pprint(list(store.execute('select * from fileinfo')))
[   (   'kgtk/tests/data/kypher/graph.tsv',
        205,
        1597353182.1801062,
        None,
        None)]
>>> cq.pp.pprint(list(store.execute('select * from graphinfo')))
[   (   'graph_1',
        None,
        "['node1', 'label', 'node2', 'id']",
        4096,
        1598648612.7562318)]

>>> store.close()
"""

"""
# Large DB times and sizes:
#
# Summary:
# - Wikidata edges file, 1.15B edges, 16GB compressed, 78GB DB size, 20min import, 4.5min analyze
# - index on node1 column, 16.5min, 22GB DB growth, 1.25min analyze
# - analyze adds about 10% run/import time for index, 20% run/import time for table
# - full 4-column index doubles DB size, increases import time by 3.3x
# Optimizations:
# - we might be able to build a covering index for 'id' to save storage for one index
# - we could use 'id' as the primary key and build a 'without rowid' table
# - we could build two-column indexes: (node1, label), (label, node2), (node2, label)
# - we might forgo analyzing tables and only do it on indexes

# Wikidata edges file (1.15B edges):
> ls -l $EDGES
-rw-r--r-- 1 hans isdstaff 16379491562 Aug 14 18:12 /data/kgtk/wikidata/run3/wikidata-20200803-all-edges.tsv.gz

# Import:
> time kgtk --debug query -i $EDGES --graph-cache /data/tmp/store.db --limit 10
IMPORT graph data into table graph_1 from /data/kgtk/wikidata/run3/wikidata-20200803-all-edges.tsv.gz
  ..............
1517.701u 167.970s 20:14.79 138.7%	0+0k 29045920+153711296io 0pf+0w

# DB size:
> ls -l /data/tmp/store.db 
-rw-r--r-- 1 hans isdstaff 78699548672 Sep 11 00:00 /data/tmp/store.db

# Analyze graph table:
> time sqlite3 /data/tmp/store.db 'analyze graph_1'
30.410u 75.243s 4:23.90 40.0%	0+0k 153709096+40io 3pf+0w

# DB size:
> ls -l /data/tmp/store.db 
-rw-r--r-- 1 hans isdstaff 78699552768 Sep 11 09:39 /data/tmp/store.db

# Index creation on node1:
> time kgtk --debug query -i $EDGES --graph-cache /data/tmp/store.db \
       --match "edge: (p:Q52353442)-[r]->(y)" \
       --limit 1000
CREATE INDEX on table graph_1 column node1
  .............
699.576u 106.269s 16:30.38 81.3%	0+0k 190371536+104441192io 3424pf+0w

# DB size:
> ls -l /data/tmp/store.db 
-rw-r--r-- 1 hans isdstaff 100584587264 Sep 11 11:15 /data/tmp/store.db

# Analyze index:
> time sqlite3 /data/tmp/store.db 'analyze graph_1_node1_idx'
68.563u 6.544s 1:15.24 99.8%	0+0k 19904088+48io 0pf+0w
"""



"""
VectorStore to support Kypher queries over embedding and other vectors.
"""

import sys
import os
import os.path
import itertools
import copy
import math

import numpy as np
import h5py
import faiss

from   kgtk.kypher.utils import *
from   kgtk.exceptions import KGTKException
import kgtk.kypher.indexspec as ispec
from   kgtk.kypher.functions import SqlFunction


### NOTES, TO DO:

# Vector storage:
# - the basic idea is that we use the database or associated file-based storage
#   to store vectors in a parsed, binary format that can then be quickly converted
#   into NumPy format at query time to compute similarities, etc.
# - during import we cut out source vector columns, parse them, and then store
#   the binary data either back in those column as BLOB values ('inline' scheme),
#   or in external NumPy files or HD5 stores.  Other auxiliary values such as
#   precomputed norms, cluster IDs, etc. are also stored in auxiliary columns.  The
#   imported data tables behave like regular KGTK graph tables, the only difference
#   is that vector data has been transformed or offloaded to external stores.
#   The query translator translates column access to appropriate function calls
#   when needed to reconstitute vectors transparently
# - 'inline'- stored data is retrieved directly like other KGTK column values,
#   the only difference is that they are transferred as byte strings which can
#   be very efficiently turned back into NumPy vectors via 'frombuffer' calls
# - externally stored data is stored in NumPy memmap files or HD5 datasets which
#   provide a transparent array interface, in this case vectors are accessed
#   by the 'rowid' of the row from which they were extracted
# - SQLite functions can only accept and return literals and byte strings, we
#   cannot pass arbitrary Python objects such as NumPy vectors.  This works fine
#   for the inline storage scheme, but for external access we have to translate
#   vector accesses into references that can then be converted to vectors by
#   the computations they are passed to (supported by new generalized function API)

# Performance:
# - we (optionally) do load-time vector normalization and store the computed norms
#   in auxiliary columns so we can later denormalize if possible; this allows us
#   to use a simple dot product to compute similarity instead of having to compute
#   the vector norms as well which is somewhat expensive
# - we use the new SQL function definition and translation API to generate vector-column
#   specific translations that have various needed values such as vector data types,
#   data set handles, etc. baked in as call-specific closure variables, so they do
#   not have to be provided or computed when those functions are called.  This keys
#   in on provided call-time arguments to infer what optimizations are possible,
#   generic computation codes cover cases where optimizations aren't possible
# - 'inline' storage is generally the fastest, NumPy provides similar import time
#   and vector access that is about 2x slower.  HD5 also provides similar import
#   time but much slower vector-at-a-time access (40x slower than 'inline')
# - in general, linking vector computations into the inner loop of SQL queries
#   prevents us from using vectorized NumPy operations which are generally faster,
#   but this seems unavoidable without completely changing the DB querying paradigm

# Indexing:
# some pointers that are possibly relevant to creating a FAISS index usable in concert
# with the DB; in particular, we'd only do the clustering, find the relevant cluster(s)
# for each vector, and then do the lookup and vector regeneration ourselves:
# - https://github.com/facebookresearch/faiss/issues/506
# - https://www.pinecone.io/learn/composite-indexes/
# - https://github.com/facebookresearch/faiss/issues/666 (need to use float32 to interface with API)


### VectorStore

class MasterVectorStore(object):
    """Manage auxiliary vector stores to store and access embeddings and other vectors.
    Each individual store manages the vectors of a single KGTK file vector column.
    """

    def __init__(self, store):
        """Create a master vector store for the SQL store 'store'.
        """
        self.store = store
        self.vector_stores = {}

    def close(self):
        for vstore in self.vector_stores.values():
            vstore.close()

    def commit(self):
        for vstore in self.vector_stores.values():
            vstore.commit()

    def get_store_key(self, table, column):
        return (table, column)
    
    def get_store_name(self, table, column):
        return f'{table}:{column}'

    def has_vector_store(self, table, column):
        """Return True if a vector store object for 'table->column' exists, None otherwise.
        """
        return self.vector_stores.get((table, column)) is not None
    
    def get_vector_store(self, table, column, index_spec=None):
        """Return a vector store object for 'table->column'.  If it does not yet exist,
        create it according to 'index_spec' (which must not be None in that case).
        """
        vstore = self.vector_stores.get((table, column))
        if vstore is not None and vstore.index_spec != index_spec and index_spec is not None:
            # redefinition with a different index spec, replace the old one:
            vstore = None
        if vstore is None:
            if index_spec is None:
                # see if this table column was previously imported with a vector index spec:
                if not self.store.is_vector_column(table, column):
                    raise KGTKException(f'{table}.{column} is not a vector column')
                index_spec = self.store.get_vector_indexes(table)
                index_spec = self.store.normalize_vector_index_specs(table, index_spec)
            if not isinstance(index_spec, ispec.VectorIndex) or not column in index_spec.index.columns:
                raise KGTKException(f'cannot create vector store without valid index spec for {table}.{column}')
            store_type = index_spec.index.columns[column].store
            if store_type == ispec.VectorIndex.STORE_INLINE:
                vstore = InlineVectorStore(self, table, column, index_spec=index_spec)
            elif store_type == ispec.VectorIndex.STORE_NUMPY:
                vstore = NumpyVectorStore(self, table, column, index_spec=index_spec)
            elif store_type == ispec.VectorIndex.STORE_HD5:
                vstore = Hd5VectorStore(self, table, column, index_spec=index_spec)
            else:
                KGTKException(f'unsupported vector store type: {store_type}')
            self.vector_stores[(table, column)] = vstore
        return vstore

    def drop_vector_store_data(self, table, column):
        """If a vector store object for 'table->column' exists, delete any of its data, no-op otherwise.
        """
        if self.has_vector_store(table, column):
            self.get_vector_store(table, column).drop_store_data()

    def get_graph_table_schema(self, table, column, schema):
        """Return a possibly modified version of 'schema' for the graph table containing
        vectors indexed and stored by the vector store for 'table->column'.
        """
        return self.get_vector_store(table, column).get_graph_table_schema(schema)

    def import_vectors(self, table, column, colidx, rows):
        return self.get_vector_store(table, column).import_vectors(colidx, rows)

    @staticmethod
    def get_vector_store_as_array(sql_store, table, column, cache=None):
        """Return an array-like handle that can be used to access the vectors of
        the vector store for 'table->column'.  'cache' can be used to cache the
        array object in SQL vector functions for fast repeated access.
        """
        master_store = sql_store.get_vector_store()
        # this requires a valid index spec to be available in 'table's graph info:
        vstore = master_store.get_vector_store(table, column)
        ds = vstore.get_store_as_array()
        if cache:
            cache[0] = ds
        return ds

    def vector_column_to_sql(self, table, column, table_alias=None):
        """Return an SQL expression that can be used to generate the actual vectors
        corresponding to the vector-indexed 'column' variable of 'table'.
        """
        return self.get_vector_store(table, column).vector_column_to_sql(table_alias=table_alias)
    
    def vector_column_to_reference_sql(self, table, column, table_alias=None):
        """Return an SQL expression that can be used to access the associated
        vector, using an optimizable vector reference for array-based stores.
        """
        return self.get_vector_store(table, column).vector_column_to_reference_sql(table_alias=table_alias)


class VectorStore(object):
    """Auxiliary vector store to store and access embeddings and other vectors.
    Each individual vector store manages the vectors of a single KGTK file vector column,
    even though multiple stores might share a single physical vector store database (e.g., for HD5).
    """

    def __init__(self, master, table, column, index_spec=None):
        """Create an auxiliary VectorStore for vector column 'table->column' and associate
        it with the master store 'master' and the 'index_spec' describing the store.
        """
        self.master = master
        self.store = master.store
        self.table = table
        self.column = column
        self.index_spec = index_spec
        self.dbfile = None
        self.vector_ndim = None
        self.vector_dtype = None
        self.vector_norm = None
        self.ntotal = None
        self.nn_index = None

    def close(self):
        raise KGTKException('not implemented')

    def commit(self):
        raise KGTKException('not implemented')

    def get_store_key(self):
        return self.master.get_store_key(self.table, self.column)
    
    def get_store_name(self):
        return self.master.get_store_name(self.table, self.column)
    
    def drop_store_data(self):
        """If a vector dataset object for 'table->column' exists, delete it, no-op otherwise.
        """
        raise KGTKException('not implemented')

    def get_store_as_array(self):
        """Return an array-like handle that can be used to access the vectors of this store.
        """
        raise KGTKException('not implemented')

    def get_graph_table_schema(self, schema):
        """Return a possibly modified version of 'schema' for the graph table containing
        vectors indexed and stored by this vector store.
        """
        return schema

    def get_norm_column_name(self):
        return f'{self.column};_kgtk_vec_norm'

    def get_quantizer_cell_column_name(self):
        return f'{self.column};_kgtk_vec_qcell'
    
    def get_vector_index_options(self):
        """Return the full options dict for the index spec of this store.
        """
        return self.index_spec.index.columns[self.column]

    def get_vector_format(self):
        return self.index_spec.index.columns[self.column].fmt
            
    def get_vector_dtype(self):
        if self.vector_dtype is None:
            self.vector_dtype = np.dtype(self.index_spec.index.columns[self.column].dtype)
        return self.vector_dtype

    def get_vector_norm(self):
        if self.vector_norm is None:
            normspec = self.index_spec.index.columns[self.column].norm
            # since we now have NN-indexes, we only normalize if explicitly requested:
            if normspec is True:
                # if just True was explicitly specified, use this as the default:
                normspec = ispec.VectorIndex.NORM_L2
            self.vector_norm = normspec
        return self.vector_norm

    def get_vector_ndim(self):
        raise KGTKException('not implemented')
            
    def get_store_ntotal(self):
        """Return the total number of vectors stored in this store.
        """
        raise KGTKException('not implemented')

    def guess_vector_format(self, example_vector, seps=(' ', ',', ';', ':', '|')):
        """Try to infer an encoding from an 'example_vector'.  For now we only handle
        text strings of numbers separated by one of 'seps' that can be parsed by numpy.
        """
        if not isinstance(example_vector, str):
            raise KGTKException(f"can only handle vectors in text format'")
        example_vector = example_vector.strip()
        for sep in seps:
            if sep in example_vector:
                return 'string', sep
        raise KGTKException(f"unhandled vector format'")

    def parse_vectors(self, vectors, fmt=None, dtype=None):
        fmt = fmt or self.get_vector_format()
        if fmt != ispec.VectorIndex.FORMAT_AUTO:
            raise KGTKException(f"cannot yet handle vector format '{fmt}'")
        dtype = dtype or self.get_vector_dtype()
        vectors = list(vectors)
        if len(vectors) == 0:
            return vectors
        fmt, sep = self.guess_vector_format(vectors[0])
        # parsing text vectors into arrays takes the most time of an import (about the same
        # as uncompressing the data with the gzip library); we tried various other things
        # np.loadtxt, parsing into list to array, json.loads, etc., but to no significant
        # improvement, so we stick with this for now:
        return [np.fromstring(vec, dtype=dtype, sep=sep) for vec in vectors]

    def normalize_vectors(self, vectors):
        """Normalize 'vectors' according to the specified norm and return the normalized
        vectors and a parallel list of norms.
        """
        norm = self.get_vector_norm()
        if norm != ispec.VectorIndex.NORM_L2:
            raise KGTKException(f"cannot handle vector norm '{norm}'")
        # TO DO: see if we are losing too much performance here by not using 2-D array ops:
        norms = [np.linalg.norm(v) for v in vectors]
        vectors = [v / n for v, n in zip(vectors, norms)]
        return vectors, norms

    def import_vectors(self, colidx, rows):
        raise KGTKException('not implemented')

    def vector_column_to_sql(self, table_alias=None):
        """Return an SQL expression that can be used to generate the actual vectors
        corresponding to the vector-indexed 'column' variable of 'table'.
        """
        raise KGTKException('not implemented')
    
    def vector_column_to_reference_sql(self, table_alias=None):
        """Return an SQL expression that can be used to access the associated
        vector, using an optimizable vector reference for array-based stores.
        """
        raise KGTKException('not implemented')

    def get_vectors_as_array(self, n, offset=0, buffer=None):
        """Return 'n' contiguous vectors of this store starting at 'offset'.
        If 'buffer is supplied, copy the vectors into it (assuming it is big enough and of the right type).
        Return the array of vectors and the number of vectors returned in it.
        """
        raise KGTKException('not implemented')
    
    def get_qcell_vectors(self, qcell):
        """Return all vectors assigned to this quantizer 'qcell' as an array.
        Return their zero-based rowid's as a second array of vector IDs.
        """
        raise KGTKException('not implemented')

    def get_vector_rows_by_id(self, ids):
        """Retrieve vector rows based on a single or collection of zero-based vector row 'ids'.
        Each returned tuple contains the 'id', 'node1', 'label', vector-id and respective vector
        of that row.  Any additional relevant columns such as norm, qcell,  etc. will also be
        returned if they are defined.
        """
        if isinstance(ids, (int, str)):
            ids = [ids]
        
        # NOTE:  DB 'rowid's are 1-based
        # TO DO: construct and cache this query, instead of redoing this every time:
        normcol = f', {sql_quote_ident(self.get_norm_column_name())}' if self.get_vector_norm() else ''
        qcellcol = self.get_quantizer_cell_column_name()
        qcellcol = f', {sql_quote_ident(qcellcol)}' if self.store.has_table_column(self.table, qcellcol) else ''
        query = f"""
            SELECT id, node1, label, {self.table}.rowid-1, {sql_quote_ident(self.column)} {normcol} {qcellcol}
            FROM {self.table}
            WHERE {self.table}.rowid = ?"""

        # this is a super-method which doesn't actually fill in vector objects, those are added below:
        # TO DO: possibly batch this instead of going id-by-id, but for now this is just used to
        # retrieve small batches of top-k results, so batching is not an issue
        rows = []
        for id in ids:
            rows += [list(row) for row in self.store.execute(query, (int(id)+1,))]
        return rows

    def get_nearest_neighbor_index(self):
        """Return an appropriate nearest neighbor index object for this store if one has been configured,
        return False otherwise.
        """
        if self.nn_index is None:
            self.nn_index = self.index_spec.index.columns[self.column].nn
            if self.nn_index:
                # for now this is all we support:
                self.nn_index = FaissIndex(self)
        return self.nn_index


class InlineVectorStore(VectorStore):
    """Auxiliary vector store that stores vectors inline as blob values in SQL rows.
    """

    def __init__(self, master, table, column, index_spec=None):
        """Create an auxiliary VectorStore for vector column 'table->column' and associate
        it with the master store 'master' and the 'index_spec' describing the store.
        """
        super().__init__(master, table, column, index_spec=index_spec)

    def close(self):
        pass

    def commit(self):
        pass

    def drop_store_data(self):
        """If a vector dataset object for 'table->column' exists, delete it, no-op otherwise.
        """
        pass

    def get_graph_table_schema(self, schema):
        """Return a possibly modified version of 'schema' for the graph table containing
        vectors indexed and stored by this vector store.
        """
        schema = copy.deepcopy(schema)
        for column, options in schema.columns.items():
            column = options._name_
            if column == self.column:
                options['type'] = 'BLOB'
                break
        if self.get_vector_norm():
            # append norm column at the end:
            norm_column = self.get_norm_column_name()
            # TO DO: if we use FLOAT here, we get back bytes, figure out why:
            schema.columns[norm_column] = sdict['_name_': norm_column, 'type': 'TEXT']
        return schema

    def get_vector_ndim(self):
        if self.vector_ndim is None:
            # for now, infer it from the first row in the table, but maybe we need to store it somewhere explicitly:
            for (vec,) in self.store.execute(f'SELECT {sql_quote_ident(self.column)} FROM {self.table} LIMIT 1'):
                self.vector_ndim = len(np.frombuffer(vec, dtype=self.get_vector_dtype()))
        return self.vector_ndim

    def get_store_ntotal(self):
        """Return the total number of vectors stored in this store.
        """
        if self.ntotal is None:
            # for now, infer it from the size of the table, but maybe we need to store it somewhere explicitly:
            self.ntotal = self.store.get_table_row_count(self.table)
        return self.ntotal

    def import_vectors(self, colidx, rows):
        """Parse a collection of source vectors stored at 'rows[colidx]' into a two-dimensional
        numpy array of vectors, and incrementally add them to this vector store.
        """
        source_vectors = map(lambda row: row[colidx], rows)
        vectors = self.parse_vectors(source_vectors)
        if len(vectors) == 0:
            return rows
        # update these values dynamically during import:
        if self.vector_ndim is None:
            self.vector_ndim = len(vectors[0])
            self.ntotal = len(vectors)
        else:
            self.ntotal += len(vectors)
        norms = None
        if self.get_vector_norm():
            vectors, norms = self.normalize_vectors(vectors)
        vectors = map(lambda v: v.tobytes(), vectors)
        any(itertools.filterfalse(lambda row: not row.__setitem__(colidx, next(vectors)), rows))
        if norms is not None:
            # append stringified norms to the end of each row - figure out why we can't use floats:
            norms = iter(norms)
            any(itertools.filterfalse(lambda row: not row.append(str(next(norms))), rows))
        return rows

    def vector_column_to_sql(self, table_alias=None):
        """Return an SQL expression that can be used to generate the actual vectors
        corresponding to the vector-indexed 'column' variable of 'table'.
        """
        table = table_alias or self.table
        return f'{table}.{self.column}'

    def vector_column_to_reference_sql(self, table_alias=None):
        """Return an SQL expression that can be used to access the associated
        vector, using an optimizable vector reference for array-based stores.
        """
        return self.vector_column_to_sql(table_alias=table_alias)
    
    def get_vectors_as_array(self, n, offset=0, buffer=None):
        """Return 'n' contiguous vectors of this store starting at 'offset'.
        If 'buffer is supplied, copy the vectors into it (assuming it is big enough and of the right type).
        Return the array of vectors and the number of vectors returned in it.
        """
        # NOTE: if a buffer is supplied, float type conversions will happen automatically
        ntotal = self.get_store_ntotal()
        end = min(offset + n, ntotal)
        nvecs = end - offset
        dtype = self.get_vector_dtype()
        if buffer is None:
            # allocate it for size=n, even if nvecs is smaller than that:
            ndim = self.get_vector_ndim()
            buffer = np.zeros(n * ndim, dtype=dtype)
            buffer.shape = (n, ndim)
        # note that 'rowid's are 1-based:
        for i, (vec,) in enumerate(self.store.execute(f"""
            SELECT {sql_quote_ident(self.column)} FROM {self.table}
            WHERE {self.table}.rowid >= ?
            LIMIT ?""", (offset + 1, n))):
            buffer[i] = np.frombuffer(vec, dtype=dtype)
        return buffer, nvecs

    def get_qcell_vectors(self, qcell):
        """Return all vectors assigned to this quantizer 'qcell' as an array.
        Return their zero-based rowid's as a second array of vector IDs.
        """
        dtype = self.get_vector_dtype()
        qcell_colname = self.get_quantizer_cell_column_name()
        vectors = []
        vectorids = []
        # NOTE: DB 'rowid's are 1-based
        # PERFORMANCE NOTE: the speed of this query greatly depends on how contiguous rows
        # with this 'qcell' ID are on disk, so sorting by qcell-ID speeds things up a lot
        for vec, vecid in self.store.execute(f"""
            SELECT {sql_quote_ident(self.column)}, {self.table}.rowid FROM {self.table}
            WHERE {self.table}.{sql_quote_ident(qcell_colname)} = ?""", (str(qcell),)):
            vectors.append(np.frombuffer(vec, dtype=dtype))
            vectorids.append(int(vecid)-1)
        if len(vectors) > 0:
            return np.vstack(vectors), np.array(vectorids, dtype=np.int64)
        else:
            return np.zeros(0, dtype=dtype), np.zeros(0, dtype=np.int64)

    def get_vector_rows_by_id(self, ids):
        """Retrieve vector rows based on a single or collection of zero-based vector row 'ids'.
        Each returned tuple contains the 'id', 'node1', 'label', vector-id and respective vector
        of that row.  Any additional relevant columns such as norm, qcell,  etc. will also be
        returned if they are defined.
        """
        rows = super().get_vector_rows_by_id(ids)
        dtype = self.get_vector_dtype()
        for row in rows:
            row[4] = np.frombuffer(row[4], dtype=dtype)
        return rows


# NumPy memmap notes:
# - see: https://numpy.org/devdocs/reference/generated/numpy.lib.format.html
# - this is significantly faster than the HD5 access (goes from 2.5 to 0.68 secs for 10k vecs)
# - further speedups down to 0.16 secs possible with vector norm precomputation
#   (see .work/embeddings-precompute-norms.log - note that the dot-product cost for this
#   is 0.015 secs, so we have another factor 10 head room for speeding up vector access)
# - np.frombuffer is really fast, 10M calls to generate a 100d array from bytes takes 4.25 secs
#   ( x = np.frombuffer(bytes, dtype=np.float32) ), so storing the bytes on the DB as 
#   binary blobs might be the best thing after all (about 20secs for 10M vectors with precomputed norms)
# - np.frombuffer takes an offset, so we could encode a norm as the first element and
#   read from an offset to skip it (but we now store the norms separately for a simpler format)
# - we could use the struct package to get a norm value encoded in the first 4 bytes:
#   - fmt=struct.Struct('<f'), fmt.unpack(vbytes[0:4]) -> (-0.10274789482355118,)

class NumpyVectorStore(VectorStore):
    """Auxiliary vector store that stores vectors via numpy memmaps.
    """

    def __init__(self, master, table, column, index_spec=None):
        """Create an auxiliary VectorStore for vector column 'table->column' and associate
        it with the master store 'master' and the 'index_spec' describing the store.
        """
        super().__init__(master, table, column, index_spec=index_spec)
        self.dbfile = None
        self.external_dbfile = None
        self.conn = None

    def get_store_dbfile(self):
        if self.dbfile is None:
            self.dbfile = self.get_store_external_file()
            if self.dbfile is None:
                vsname = self.get_store_name()
                self.dbfile = f'{self.store.dbfile}.vec.{vsname}.npy'
        return self.dbfile

    def has_external_file(self):
        """Return True if this store is defined by an external .npy file.
        """
        if self.dbfile is None:
            self.dbfile = self.get_store_external_file()
        return self.external_dbfile is not None

    def get_store_conn(self, mode='r', mmap=False):
        """Return the dataset file connection object for this dataset, open it if necessary
        according to 'mode'.  If it is already open in read mode but a write-mode is
        requested, close it and reopen it according to the new mode (unless the underlying
        SQL store is in read-only mode).
        """
        conn = self.conn
        if conn is None:
            dbfile = self.get_store_dbfile()
            if self.store.readonly or self.has_external_file():
                # ignore mode, always open as read-only:
                mode = 'r'
            elif not os.path.exists(dbfile):
                # create an empty .npy file so we can open it with any mode, we'll fix up info later:
                np.save(dbfile, np.zeros(0))
            if mmap:
                conn = np.load(dbfile, mmap_mode='r', fix_imports=False)
            else:
                mode += 'b' if 'b' not in mode else ''
                conn = open(dbfile, mode)
            self.conn = conn
        elif mode != 'r' and conn.mode == 'r' and not isinstance(conn, np.memmap) and not self.store.readonly:
            # regular file opened as read-only, close and reopen with provided write mode:
            conn.close()
            conn = open(dbfile, mode)
            self.conn = conn
        return conn

    def get_store_as_array(self):
        """Return an array-like handle that can be used to access the vectors of this store.
        """
        return self.get_store_conn(mode='r', mmap=True)

    def read_dataset_header(self, dsspec):
        try:
            if isinstance(dsspec, str):
                ds = open(dsspec, 'rb')
            else:
                ds = dsspec
            ds.seek(0)
            version = np.lib.format.read_magic(ds)[0]
            if version == 1:
                shape, fo, dt = np.lib.format.read_array_header_1_0(ds)
            else:
                shape, fo, dt = np.lib.format.read_array_header_2_0(ds)
            return shape, fo, dt, version
        finally:
            if isinstance(dsspec, str):
                ds.close()

    def update_dataset_header(self, ds, ndim=None, dtype=None):
        dtype = dtype or self.get_vector_dtype()
        curpos = ds.tell()
        shape, fo, dt, version = self.read_dataset_header(ds)
        startpos = ds.tell()
        if shape[0] == 0:
            if ndim is None:
                raise KGTKException(f"cannot determine ndim of vector dataset")
        else:
            _, ndim = shape
        ds.seek(0, 2)
        endpos = ds.tell()
        dtype = np.dtype(dtype)
        nbytes = dtype.itemsize
        count = (endpos - startpos) // (ndim * nbytes)
        desc = {'descr': np.lib.format.dtype_to_descr(dtype), 'fortran_order': fo, 'shape': (count, ndim)}
        ds.seek(0)
        if version == 1:
            np.lib.format.write_array_header_1_0(ds, desc)
        else:
            np.lib.format.write_array_header_2_0(ds, desc)
        ds.seek(curpos)

    def close(self):
        if self.conn is not None:
            if not isinstance(self.conn, np.memmap):
                self.conn.close()
        self.conn = None

    def commit(self):
        self.close()

    def drop_store_data(self):
        """If a vector dataset object for 'table->column' exists, delete it, no-op otherwise.
        """
        self.close()
        dbfile = self.get_store_dbfile()
        if not self.has_external_file() and os.path.exists(dbfile):
            os.remove(dbfile)

    def get_graph_table_schema(self, schema):
        """Return a possibly modified version of 'schema' for the graph table containing
        vectors indexed and stored by this vector store.
        """
        schema = copy.deepcopy(schema)
        if self.get_vector_norm() and not self.has_external_file():
            # append norm column at the end:
            norm_column = self.get_norm_column_name()
            # TO DO: if we use FLOAT here, we get back bytes, figure out why:
            schema.columns[norm_column] = sdict['_name_': norm_column, 'type': 'TEXT']
        return schema

    def get_vector_ndim(self):
        if self.vector_ndim is None:
            dbfile = self.get_store_dbfile()
            shape, fo, dt, version = self.read_dataset_header(dbfile)
            if shape[0] == 0:
                raise KGTKException(f"cannot determine shape of vector dataset")
            self.ntotal = shape[0]
            self.vector_ndim = shape[1]
        return self.vector_ndim

    def get_vector_norm(self):
        if self.vector_norm is None:
            if self.has_external_file() and self.index_spec.index.columns[self.column].norm is None:
                # this is different than the default behavior if we have an external store file,
                # in which case we use False as the default if the user didn't specify anything,
                # since we can't normalize the vectors on the external file; however, we still
                # want to be able to indicate normalization if the external vectors are normalized:
                self.vector_norm = False
            else:
                return super().get_vector_norm()
        return self.vector_norm

    def get_store_ntotal(self):
        """Return the total number of vectors stored in this store.
        """
        if self.ntotal is None:
            # this will assign it as a side-effect:
            self.get_vector_ndim()
        return self.ntotal

    def get_store_external_file(self):
        if self.dbfile is None:
            self.external_dbfile = self.index_spec.index.columns[self.column].ext
            self.dbfile = self.external_dbfile
            if self.dbfile is not None:
                shape, fo, dt, version = self.read_dataset_header(self.dbfile)
                # TO DO: maybe check for mismatches from the index spec, for now
                # we simply accept whatever comes with the external NumPy file:
                self.ntotal = shape[0]
                self.vector_ndim = shape[1]
                self.dtype = np.dtype(dt)
        return self.dbfile
    
    def import_vectors(self, colidx, rows):
        """Parse a collection of source vectors stored at 'rows[colidx]' into a two-dimensional
        numpy array of vectors, and incrementally add them to this vector store.
        """
        if self.has_external_file():
            # no import required, but just in case, still set the vector column fields to '':
            any(itertools.filterfalse(lambda row: not row.__setitem__(colidx, ''), rows))
            return rows
        # parse and import the vectors:
        source_vectors = map(lambda row: row[colidx], rows)
        vectors = self.parse_vectors(source_vectors)
        if len(vectors) == 0:
            return rows
        ds = self.conn
        if ds is None or 'a' not in ds.mode:
            # maybe use a 'create_dataset' here for a cleaner API
            # (we do need 'r+' here so we can append and read/write the header):
            ds = self.get_store_conn(mode='r+')
            # a+ doesn't do what we want, so we have to position by hand:
            ds.seek(0, 2)
        norms = None
        if self.get_vector_norm():
            vectors, norms = self.normalize_vectors(vectors)
        # writing the whole chunk in a single write op doesn't save us anything:
        for vec in vectors:
            ds.write(vec.tobytes())
        self.update_dataset_header(ds, ndim=len(vectors[0]))
        # set all the respective fields in the KGTK source file tuples to the empty value:
        any(itertools.filterfalse(lambda row: not row.__setitem__(colidx, ''), rows))
        if norms is not None:
            # append stringified norms to the end of each row - figure out why we can't use floats:
            norms = iter(norms)
            any(itertools.filterfalse(lambda row: not row.append(str(next(norms))), rows))
        return rows

    def vector_column_to_sql(self, table_alias=None):
        """Return an SQL expression that can be used to generate the actual vectors
        corresponding to the vector-indexed 'column' variable of 'table'.
        """
        vfn = SqlFunction.get_function('_kvec_get_vector', store=self.store, uniq=True, xstore=self)
        vfn.load()
        table = table_alias or self.table
        return f'{vfn.get_name()}({table}.rowid)'

    def vector_column_to_reference_sql(self, table_alias=None):
        """Return an SQL expression that can be used to access the associated
        vector, using an optimizable vector reference for array-based stores.
        """
        table = table_alias or self.table
        # NOTE: rowid is 1-based
        return f'{table}.rowid'

    def get_vectors_as_array(self, n, offset=0, buffer=None):
        """Return 'n' contiguous vectors of this store starting at 'offset'.
        If 'buffer is supplied, copy the vectors into it (assuming it is big enough and of the right type).
        Return the array of vectors and the number of vectors returned in it.
        """
        # NOTE: if a buffer is supplied, float type conversions will happen automatically
        ntotal = self.get_store_ntotal()
        end = min(offset + n, ntotal)
        nvecs = end - offset
        if buffer is not None:
            # caller must ensure that buffer is large enough:
            buffer[0:nvecs] = self.get_store_as_array()[offset:end]
            return buffer, nvecs
        else:
            return self.get_store_as_array()[offset:end], nvecs

    def get_qcell_vectors(self, qcell):
        """Return all vectors assigned to this quantizer 'qcell' as an array.
        Return their zero-based rowid's as a second array of vector IDs.
        """
        dtype = self.get_vector_dtype()
        qcell_colname = self.get_quantizer_cell_column_name()
        store_vectors = self.get_store_as_array()
        vectors = []
        vectorids = []
        # NOTE: 'rowid's are 1-based
        # PERFORMANCE NOTE: the speed of this query greatly depends on how contiguous rows
        # with this 'qcell' ID are on disk, so sorting by qcell-ID speeds things up a lot
        for (vecid,) in self.store.execute(f"""
            SELECT {self.table}.rowid FROM {self.table}
            WHERE {self.table}.{sql_quote_ident(qcell_colname)} = ?""", (str(qcell),)):
            vecid = int(vecid)-1
            vectors.append(store_vectors[vecid])
            vectorids.append(vecid)
        if len(vectors) > 0:
            return np.vstack(vectors), np.array(vectorids, dtype=np.int64)
        else:
            return np.zeros(0, dtype=dtype), np.zeros(0, dtype=np.int64)

    def get_vector_rows_by_id(self, ids):
        """Retrieve vector rows based on a single or collection of zero-based vector row 'ids'.
        Each returned tuple contains the 'id', 'node1', 'label', vector-id and respective vector
        of that row.  Any additional relevant columns such as norm, qcell,  etc. will also be
        returned if they are defined.
        """
        rows = super().get_vector_rows_by_id(ids)
        store_vectors = self.get_store_as_array()
        for row in rows:
            row[4] = store_vectors[row[3]]
        return rows


class Hd5VectorStore(VectorStore):
    """Auxiliary vector store to store and access embeddings and other vectors.
    """

    def __init__(self, master, table, column, index_spec=None):
        """Create an auxiliary VectorStore for vector column 'table->column' and associate
        it with the master store 'master' and the 'index_spec' describing the store.
        """
        super().__init__(master, table, column, index_spec=index_spec)
        # a vector cache is tightly linked to its parent SQL store, so we use a fixed pattern here:
        self.dbfile = self.store.dbfile + '.vec.hdf5'
        self.conn = None

    def get_linked_stores(self):
        # all HD5 stores of this master store share a single DB file and connection:
        return [vstore for vstore in self.master.vector_stores.values() if isinstance(vstore, Hd5VectorStore)]

    def get_conn(self, mode='r'):
        """Return the HD5 file connection object for this store, open it if necessary
        according to 'mode'.  If it is already open in read mode but a write-mode is
        requested, close it and reopen it according to the new mode (unless the underlying
        SQL store is in read-only mode).  We only upgrade the mode from 'r' to 'r+' if
        needed, we never go the other direction.
        """
        if self.conn is None:
            for vstore in self.get_linked_stores():
                if vstore.conn is not None:
                    self.conn = vstore.conn
                    break
        if self.conn is None:
            if self.store.readonly:
                # ignore mode, always open as read-only:
                mode = 'r'
            elif mode == 'r' and not os.path.exists(self.dbfile):
                # create the file so we can open it with 'r':
                h5py.File(self.dbfile, 'a').close()
            self.conn = h5py.File(self.dbfile, mode)
        elif mode != 'r' and self.conn.mode == 'r' and not self.store.readonly:
            # opened as read-only, close and reopen with provided write mode:
            self.close()
            self.conn = h5py.File(self.dbfile, mode)
        return self.conn

    def close(self):
        for vstore in self.get_linked_stores():
            if vstore.conn is not None:
                vstore.conn.close()
            vstore.conn = None

    def commit(self):
        self.close()

    def get_store_dataset(self):
        """Return a vector dataset object for 'table->column' if it exists, None otherwise.
        """
        return self.get_conn().get(self.get_store_name())

    def get_store_as_array(self):
        """Return an array-like handle that can be used to access the vectors of this store.
        """
        return self.get_store_dataset()

    def drop_store_data(self):
        """If a vector dataset object for 'table->column' exists, delete it, no-op otherwise.
        """
        ds = self.get_store_dataset()
        if ds is not None:
            del self.get_conn(mode='r+')[self.get_store_name()]

    def get_graph_table_schema(self, schema):
        """Return a possibly modified version of 'schema' for the graph table containing
        vectors indexed and stored by this vector store.
        """
        schema = copy.deepcopy(schema)
        if self.get_vector_norm():
            # append norm column at the end:
            norm_column = self.get_norm_column_name()
            # TO DO: if we use FLOAT here, we get back bytes, figure out why:
            schema.columns[norm_column] = sdict['_name_': norm_column, 'type': 'TEXT']
        return schema

    def get_vector_ndim(self):
        if self.vector_ndim is None:
            self.ntotal, self.vector_ndim = self.get_store_as_array().shape
        return self.vector_ndim
    
    def get_store_ntotal(self):
        """Return the total number of vectors stored in this store.
        """
        if self.ntotal is None:
            # this will assign it as a side-effect:
            self.get_vector_ndim()
        return self.ntotal

    def import_vectors(self, colidx, rows):
        """Parse a collection of source vectors stored at 'rows[colidx]' into a two-dimensional
        numpy array of vectors, and incrementally add them to this vector store.
        """
        source_vectors = map(lambda row: row[colidx], rows)
        vectors = self.parse_vectors(source_vectors)
        if len(vectors) == 0:
            return rows
        ds = self.get_store_dataset()
        if ds is None:
            dsname = self.get_store_name()
            ndim = len(vectors[0])
            dtype = self.get_vector_dtype()
            chunksize = self.store.VECTOR_IMPORT_CHUNKSIZE
            conn = self.get_conn(mode='r+')
            # we need chunking so we can resize, otherwise we'd have to know the number of vectors ahead of time;
            # we use auto-chunking for now, but maybe supplying an explicit shape would be better:
            ds = conn.create_dataset(dsname, dtype=dtype, shape=(0,ndim), maxshape=(None,ndim), chunks=True)
        norms = None
        if self.get_vector_norm():
            vectors, norms = self.normalize_vectors(vectors)
        size, ndim = ds.shape
        ds.resize(size + len(vectors), axis=0)
        ds[size:] = vectors
        # set all the respective fields in the KGTK source file tuples to the empty value:
        any(itertools.filterfalse(lambda row: not row.__setitem__(colidx, ''), rows))
        if norms is not None:
            # append stringified norms to the end of each row - figure out why we can't use floats:
            norms = iter(norms)
            any(itertools.filterfalse(lambda row: not row.append(str(next(norms))), rows))
        return rows

    def vector_column_to_sql(self, table_alias=None):
        """Return an SQL expression that can be used to generate the actual vectors
        corresponding to the vector-indexed 'column' variable for 'self.table'.
        """
        vfn = SqlFunction.get_function('_kvec_get_vector', store=self.store, xstore=self)
        vfn.load()
        table = table_alias or self.table
        return f'{vfn.get_name()}({table}.rowid)'

    def vector_column_to_reference_sql(self, table_alias=None):
        """Return an SQL expression that can be used to access the associated
        vector, using an optimizable vector reference for array-based stores.
        """
        table = table_alias or self.table
        # NOTE: rowid is 1-based
        return f'{table}.rowid'

    def get_vectors_as_array(self, n, offset=0, buffer=None):
        """Return 'n' contiguous vectors of this store starting at 'offset'.
        If 'buffer is supplied, copy the vectors into it (assuming it is big enough and of the right type).
        Return the array of vectors and the number of vectors returned in it.
        """
        # NOTE: if a buffer is supplied, float type conversions will happen automatically
        ntotal = self.get_store_ntotal()
        end = min(offset + n, ntotal)
        nvecs = end - offset
        if buffer is not None:
            # caller must ensure that buffer is large enough:
            buffer[0:nvecs] = self.get_store_as_array()[offset:end]
            return buffer, nvecs
        else:
            return self.get_store_as_array()[offset:end], nvecs

    def get_qcell_vectors(self, qcell):
        """Return all vectors assigned to this quantizer 'qcell' as an array.
        Return their zero-based rowid's as a second array of vector IDs.
        """
        dtype = self.get_vector_dtype()
        qcell_colname = self.get_quantizer_cell_column_name()
        store_vectors = self.get_store_as_array()
        vectors = []
        vectorids = []
        # NOTE: 'rowid's are 1-based
        # PERFORMANCE NOTE: the speed of this query greatly depends on how contiguous rows
        # with this 'qcell' ID are on disk, so sorting by qcell-ID speeds things up a lot
        for (vecid,) in self.store.execute(f"""
            SELECT {self.table}.rowid FROM {self.table}
            WHERE {self.table}.{sql_quote_ident(qcell_colname)} = ?""", (str(qcell),)):
            vecid = int(vecid)-1
            vectors.append(store_vectors[vecid])
            vectorids.append(vecid)
        if len(vectors) > 0:
            return np.vstack(vectors), np.array(vectorids, dtype=np.int64)
        else:
            return np.zeros(0, dtype=dtype), np.zeros(0, dtype=np.int64)

    def get_vector_rows_by_id(self, ids):
        """Retrieve vector rows based on a single or collection of zero-based vector row 'ids'.
        Each returned tuple contains the 'id', 'node1', 'label', vector-id and respective vector
        of that row.  Any additional relevant columns such as norm, qcell,  etc. will also be
        returned if they are defined.
        """
        rows = super().get_vector_rows_by_id(ids)
        store_vectors = self.get_store_as_array()
        for row in rows:
            row[4] = store_vectors[row[3]]
        return rows


### Indexing:

# Metric notes:
# - from https://github.com/facebookresearch/faiss/blob/main/faiss/MetricType.h:
#     Most algorithms support both inner product and L2, with the flat
#     (brute-force) indices supporting additional metric types for vector comparison
#
#     faiss.METRIC_INNER_PRODUCT = 0   # maximum inner product search
#     faiss.METRIC_L2 = 1              # squared L2 search
#     faiss.METRIC_L1                  # L1 (aka cityblock), Manhattan-distance
#
# - stored in 'index.metric_type'
# - for Kmeans the default is squared L2 (so we don't need sqrt)
# - inner product is supported, but that might be potentially bad for quantization,
#   since now all vectors are mapped onto a unit circle (think about that)
#
# Indexing notes:
# - clustering.train method takes init centroids as a parameter, so training
#   on smaller batches of data is possible by restaring with the clusters
#   from the prior batch
#   - see https://github.com/facebookresearch/faiss/issues/531
# - see this for details on index factory:
#   https://github.com/facebookresearch/faiss/wiki/The-index-factory
#   https://gist.github.com/mdouze/c3111d5f12d1308f5adf78dcd48cdf37  (factory_string.bnf)
#   https://github.com/facebookresearch/faiss/blob/main/faiss/index_factory.cpp
# - we might be able to use faiss.IndexShards to search over multiple shards in parallel
#
# Scaling, sampling, dimensioning:
# - FAISS requires 39 min_points_per_centroid for training (otherwise warning),
#   and subsamples data if we have more than 256 max_points_per_centroid, so
#   we should mirror / approximate that when we read training data from the DB
# - to allow training on data where the required sample does not fit into RAM
#   we can use two basic approaches:
#   (1) train an index an iteration at a time where at each iteration we provide
#       a different window on the training sample that can be loaded into RAM.
#       This seems to work, a sliding window can also be used but might not help.
#       The advantage of this that we end up with a single quantizer for all
#       of the data which makes access simpler.
#   (2) train indexes on subsections of the data whose training sample fits into RAM;
#       then search on all of those during query time and combine results, or
#       use a faiss.IndexShards that does the combined searching for us.  We
#       kind of figured out how to do that, there are still a couple of wrinkles
#       to iron out.  The advantage of that is that the training of the shards
#       is on complete data slices, and that the querying is automatically threaded.
# - either way, given a max-RAM target and a set of vectors, we can compute whether
#   we need one or more training slices, and whether they need to be subsampled
# - Example situations:
#   - 55M 100d vectors, 8K vectors per centroid, 8GB of RAM
#     need 6714 centroids, pick 8192 (next power of 2),
#     min-train = 39*8192 = 319488, avg-train = 128*8192 = 1048576, max-train = 2097152,
#     min-RAM = 121MB, max-RAM = 800MB; no batching/sharding needed, train in single run,
#     even if we go to 1K vectors per centroid, we need at most 6.5GB or so
#   - 100M 1024d vectors, 1K vectors per centroid, 16GB of RAM
#     need 97657 centroids, pick 131072 (next power of 2),
#     min-train = 5,111,808, avg-train = 16,777,216, max-train = 33,554,432
#     min-RAM = 19.9GB, max-RAM = 131GB, so we need 2 10GB batches at min-data or
#     9 14.5GB batches to train at max-data with the available RAM
# - additional vector index specifications facets:
#   - ram=16GB, nlist=None (use 8K size to compute if not provided), nlist=8, niter=25


class NearestNeighborIndex(object):
    """Index object that speeds up nearest-neighbor search operations over the vectors
    in a vector store.
    """

    def __init__(self, store):
        """Initialize a nearest neighbor index for the vector store 'store'.
        """
        self.vector_store = store
        self.sql_store = self.vector_store.store

    def is_trained(self):
        """Return True if this index has a trained quantizer.
        """
        raise KGTKException('not implemented')

    def create(self, force=False, explain=False):
        """Top-level entry point that builds the index for the current set of vectors in the vector store,
        unless it has already been built previously, or if 'force' is True.  Depending on the respective
        index spec parameters, this may take a long time to compute.  If 'explain' is True, only describe
        the steps that would be taken, but do not actually build the index.
        """
        raise KGTKException('not implemented')
    
    def drop(self, explain=False):
        """Top-level entry point that deletes the index for the current set of vectors in the vector store.
        If 'explain' is True, only describe the steps that would be taken, but do not actually drop the index.
        """
        # no-op by default:
        pass
    
    def search(self, vectors, k=1, nprobe=None):
        """Search this index for the 'k' nearest neighbors of 'vectors' and return
        the result as a tuple D, V (where D are distances and V vector IDs).
        'nprobe' controls how deep the index will be searched.
        """
        raise KGTKException('not implemented')

    
class FaissIndex(NearestNeighborIndex):
    """Nearest-neighbor index implemented via calls to the FAISS library.
    The current implementation basically creates a database-centered
    faiss.IndexIVFFlat index that creates a small quantizer which gets
    loaded into memory and then very efficiently loads previously quantized
    vectors from the DB during search.  This basically creates a disk-based
    FAISS index which allows us to create, train and search very large indexes
    without having to load all vectors into RAM.  While FAISS has some limited
    and somewhat crufty on-disk support, it still requires large amounts of
    RAM to train such indexes, and the vectors are stored as part of the
    index which makes our Kypher interactions with them more difficult.
    """

    FAISS_FLOAT_TYPE = np.float32    # uniform float type used by FAISS C++ API
    FAISS_VECTOR_ID_TYPE = np.int64  # type for numeric vector IDs
    DEFAULT_CENTROID_SIZE = 8192
    DEFAULT_MIN_POINTS_PER_CENTROID = 39
    DEFAULT_MAX_POINTS_PER_CENTROID = 256
    DEFAULT_MAX_ALLOWED_RAM = 16 * 1024 * 1024 * 1024
    DEFAULT_NITER = 10
    DEFAULT_NPROBE = 8
    
    def __init__(self, store):
        """Initialize a FAISS nearest neighbor index for the vector store 'store'.
        """
        super().__init__(store)
        self.index_options = None
        self.vector_ndim = None
        self.vector_dtype = None
        self.vector_norm = None
        self.ntotal = None
        self.nlist = None
        self.nbatches = None
        self.batch_size = None
        self.niter = None
        self.nprobe = None
        self.data_buffer = None
        self.clustering = None
        self.quantizer = None
        self.search_index = None
        self.search_index_qcells = None
        self.compute_parameters()

    def compute_parameters(self):
        """Initialize any undefined parameters based on the linked vector store.
        """
        vstore = self.vector_store
        if self.index_options is None:
            self.index_options = vstore.get_vector_index_options()
        if self.vector_ndim is None:
            self.vector_ndim = vstore.get_vector_ndim()
        if self.vector_dtype is None:
            self.vector_dtype = vstore.get_vector_dtype()
        if self.vector_norm is None:
            self.vector_norm = vstore.get_vector_norm()
        if self.ntotal is None:
            self.ntotal = vstore.get_store_ntotal()
        if self.nlist is None:
            self.nlist = self.index_options.get('nlist')
            if self.nlist is None:
                # pick closest power of 2 based on default centroid size:
                self.nlist = 1 << round(math.log2(self.ntotal / self.DEFAULT_CENTROID_SIZE))
        if self.niter is None:
            self.niter = self.index_options.get('niter') or self.DEFAULT_NITER
        if self.nprobe is None:
            self.nprobe = self.index_options.get('nprobe') or self.DEFAULT_NPROBE
        if self.batch_size is None:
            nbytes = np.zeros(1, dtype=self.FAISS_FLOAT_TYPE).nbytes
            min_ram = self.vector_ndim * nbytes * self.nlist * self.DEFAULT_MIN_POINTS_PER_CENTROID
            max_ram = self.vector_ndim * nbytes * self.nlist * self.DEFAULT_MAX_POINTS_PER_CENTROID
            max_allowed_ram = self.index_options.get('ram') or self.DEFAULT_MAX_ALLOWED_RAM
            max_allowed_ram *= 0.8 # safety cushion for other buffers, etc.
            if max_ram <= max_allowed_ram:
                # single batch, pick max data size that doesn't trigger FAISS subsampling:
                self.batch_size = min(self.nlist * self.DEFAULT_MAX_POINTS_PER_CENTROID, self.ntotal)
                self.nbatches = 1
            elif min_ram <= max_allowed_ram:
                # single batch, pick as many vectors as we can fit into available RAM:
                self.batch_size = max_allowed_ram / ( self.vector_ndim * nbytes )
                self.nbatches = 1
            else:
                # multi-batch training required which likely means that we have a lot of data,
                # so train based on the minimum number of data points recommended by FAISS:
                self.nbatches = math.ceil(min_ram / max_allowed_ram)
                self.batch_size = self.nlist * self.DEFAULT_MIN_POINTS_PER_CENTROID // self.nbatches
       
    def ensure_float_type(self, array):
        """Coerce 'array' to float32 if necessary, which is the float type the FAISS API expects.
        """
        if array.dtype != self.FAISS_FLOAT_TYPE:
            return array.astype(self.FAISS_FLOAT_TYPE)
        else:
            return array

    def ensure_vector_id_type(self, array):
        """Coerce 'array' to int64 if necessary, which is the vector ID type the FAISS API expects.
        """
        if array.dtype != self.FAISS_VECTOR_ID_TYPE:
            return array.astype(self.FAISS_VECTOR_ID_TYPE)
        else:
            return array

    def ensure_vector_array(self, array):
        """Make sure 'array' is a 2-d array of vectors (in case only a single vector was provided).
        """
        if len(array.shape) == 2:
            return array
        elif len(array.shape) == 1:
            return array.reshape((1, array.shape[0]))
        else:
            raise KGTKException('cannot reshape array into 2-d form')
        
    def get_data_buffer(self):
        """Allocate a data buffer that can be used to hold self.batch_size vectors.
        """
        if self.data_buffer is None:
            self.data_buffer = np.zeros(self.batch_size * self.vector_ndim, dtype=self.FAISS_FLOAT_TYPE)
            self.data_buffer.shape = (self.batch_size, self.vector_ndim)
        return self.data_buffer

    def get_quantizer_file(self, shardid=''):
        """Return a file name that can be used to store the trained quantizer index.
        """
        # TO DO: consider storing the quantizer as a BLOB object in the database
        vstore = self.vector_store
        vsname = vstore.get_store_name()
        if shardid and not shardid.endswith('.'):
            shardid += '.'
        return f'{vstore.store.dbfile}.vec.{vsname}.faiss.{shardid}idx'

    def get_quantizer(self, error=True):
        """Return the quantizer for this vector store, raise an error if it doesn't exist
        unless 'error' is False.
        """
        if self.quantizer is None:
            qfile = self.get_quantizer_file()
            if os.path.exists(qfile):
                self.quantizer = faiss.read_index(qfile)
            elif error:
                raise KGTKException(f"cannot find quantizer for this vector store: {qfile}")
        return self.quantizer

    def is_trained(self):
        """Return True if this index has a trained quantizer.
        """
        return self.get_quantizer(error=False) is not None and self.quantizer.is_trained

    def train_quantizer_single_batch(self):
        # this could be run as a special case of multi-batch, but just in case it is
        # more advantageous to do all iterations at once, we implement this separately:
        clus = faiss.Clustering(self.vector_ndim, self.nlist)
        clus.niter = self.niter
        clus.verbose = self.sql_store.loglevel > 0
        if self.vector_norm == ispec.VectorIndex.NORM_L2:
            clus.index = faiss.IndexFlatIP(self.vector_ndim)
        else:
            clus.index = faiss.IndexFlatL2(self.vector_ndim)
        self.clustering = clus
        self.quantizer = clus.index
        vstore = self.vector_store
        buffer = self.get_data_buffer()
        # TO DO: support random sampling of batches
        buffer, nvec = vstore.get_vectors_as_array(self.batch_size, offset=0, buffer=buffer)
        self.sql_store.log(1, f'Training vector store quantizer:')
        clus.train(buffer[0:nvec], clus.index)
        return clus

    def train_quantizer_multi_batch(self):
        clus = faiss.Clustering(self.vector_ndim, self.nlist)
        clus.niter = 1
        clus.verbose = self.sql_store.loglevel > 0
        if self.vector_norm == ispec.VectorIndex.NORM_L2:
            clus.index = faiss.IndexFlatIP(self.vector_ndim)
        else:
            clus.index = faiss.IndexFlatL2(self.vector_ndim)
        self.clustering = clus
        self.quantizer = clus.index
        vstore = self.vector_store
        buffer = self.get_data_buffer()
        # TO DO: figure out whether we should use the 'nredo' parameter for this instead:
        for i in range(self.niter):
            cursor = 0
            self.sql_store.log(1, f'Training vector store quantizer iteration {i+1}:')
            for b in range(self.nbatches):
                # TO DO: support random sampling of batches, for the multi-batch case we also
                #        need to ensure that we get the same sampled batches in each iteration
                buffer, nvec = vstore.get_vectors_as_array(self.batch_size, offset=cursor, buffer=buffer)
                clus.train(buffer[0:nvec], clus.index)
                cursor += nvec
        return clus

    def train_quantizer(self, explain=False):
        """Train and save a quantizer index for this vector store.
        If 'explain' just explain what is going to be done without doing it.
        """
        if explain:
            # for now, elaborate description of process:
            self.sql_store.log(0, 'Training quantizer...')
            return
        if self.nbatches == 1:
            self.train_quantizer_single_batch()
        else:
            self.train_quantizer_multi_batch()
        faiss.write_index(self.quantizer, self.get_quantizer_file())

    def quantize_vectors(self, explain=False):
        """Quantize the vectors in this vector store with the previously trained quantizer.
        This will quantize each vector in the database and associate it with its qcell in
        its corresponding graph table.  If 'explain' just explain what is going to be done
        without doing it.
        """
        quantizer = self.get_quantizer()
        temp_table = f'_qcell_temp_{abs(hash(self.quantizer))}'
        temp_schema = sdict['_name_': temp_table, 'columns': sdict['qcell': sdict['_name_': 'qcell', 'type': 'TEXT']], 'temporary': True]
        sstore = self.sql_store
        vstore = self.vector_store
        vstable = vstore.table
        sstore.execute(sstore.get_table_definition(temp_schema))
        # TO DO: ensure that we always have a large enough batch size for this:
        buffer = self.get_data_buffer()
        cursor = 0
        sstore.log(1, f'Quantizing {self.ntotal} vectors in batches of size {self.batch_size}...')
        while cursor < self.ntotal:
            vecs, nvec = vstore.get_vectors_as_array(self.batch_size, offset=cursor, buffer=buffer)
            cursor += nvec
            D, V = quantizer.search(vecs[0:nvec], 1)
            sstore.executemany(f'INSERT INTO {temp_table} VALUES (?)', map(lambda x: (str(x[0]),), V))
            if sstore.loglevel > 0:
                sys.stderr.write('.')
                sys.stderr.flush()
        if sstore.loglevel > 0:
            sys.stderr.write('\n')
        sstore.log(1, f'Adding quantization index to database...')
        qcell_colname = vstore.get_quantizer_cell_column_name()
        qcell_index_spec = ispec.TableIndex(vstable, qcell_colname)
        header = sstore.get_table_header(vstable)
        if qcell_colname not in header:
            sstore.execute(f'ALTER TABLE {vstable} ADD COLUMN {sql_quote_ident(qcell_colname)} TEXT')
        elif sstore.has_graph_index(vstable, qcell_index_spec):
            # we have an index from a prior quantization, delete it since we need to rebuild it from scratch:
            sstore.drop_graph_index(vstable, qcell_index_spec)
        sstore.execute(
            f"""UPDATE {vstable}
                SET {sql_quote_ident(qcell_colname)} = {temp_table}.qcell
                FROM {temp_table}
                WHERE {vstable}.rowid = {temp_table}.rowid""")
        self.sort_vectors()
        sstore.ensure_graph_index(vstable, qcell_index_spec, commit=False)

    def sort_vectors(self):
        """Sort the vectors in the associated vector store by q-cell ID for improved locality during search.
        """
        # TO DO: this should become a method on vector stores
        sstore = self.sql_store
        vstore = self.vector_store
        if isinstance(vstore, InlineVectorStore):
            qcell_colname = vstore.get_quantizer_cell_column_name()
            sstore.sort_table(vstore.table, f'CAST({sql_quote_ident(qcell_colname)} AS INT)')
        else:
            sstore.log(0, 'WARNING: sorting on this type of vector store not yet implemented')

    def create(self, force=False, explain=False):
        """Top-level entry point that builds the index for the current set of vectors in the vector store,
        unless it has already been built previously, or if 'force' is True.  Depending on the respective
        index spec parameters, this may take a long time to compute.  If 'explain' is True, only describe
        the steps that would be taken, but do not actually build the index.
        """
        if not self.is_trained() or force:
            self.train_quantizer(explain=explain)
            self.quantize_vectors(explain=explain)

    def drop(self, explain=False):
        """Top-level entry point that deletes the index for the current set of vectors in the vector store.
        If 'explain' is True, only describe the steps that would be taken, but do not actually drop the index.
        """
        loglevel = 0 if explain else 1
        if self.is_trained():
            qfile = self.get_quantizer_file()
            self.sql_store.log(loglevel, f"DELETING NN-index quantizer file '{qfile}'")
            if not explain:
                # we only remove the quantizer file for now, we do not also remove the qcell
                # column, since it is harmless and will most likely be replaced with a new one:
                os.remove(qfile)
                # reinitialize to remove any stale cached information:
                self.__init__(self.vector_store)

    def get_search_index(self):
        """Return a FAISS index that can be used to search a query vector against all vectors from relevant q-cells.
        This index is created dynamically during a query, linked to a pre-trained quantizer and then quickly populated
        with relevant vectors retrieved from the database using the 'index.add_core' method.  It uses a fast,
        multi-threaded, heap-based search implementation by FAISS which beats anything we might create from scratch.
        """
        if self.search_index == None:
            quantizer = self.get_quantizer()
            self.search_index = faiss.IndexIVFFlat(quantizer, self.vector_ndim, self.nlist, quantizer.metric_type)
            self.search_index.is_trained = True
            self.search_index.verbose = self.sql_store.loglevel >= 2
            self.search_index_qcells = set()
        return self.search_index

    def get_search_index_for_qcells(self, qcells):
        """Return a search index that contains (at least) all vectors of the identified 'qcells'.
        """
        # TO DO: make this a size-limited index, currently we are reusing a search index that's already been
        # created and add to it what is missing which speeds repeat queries but might eventually run out of RAM;
        # this also changes the search quality over the course of a query, so think about this some more:
        index = self.get_search_index()
        loaded_qcells = self.search_index_qcells
        vstore = self.vector_store
        for row in self.ensure_vector_array(qcells):
            for qcell in row:
                # -1 is a FAISS code for "not found":
                if qcell >= 0 and qcell not in loaded_qcells:
                    vecs, vecids = vstore.get_qcell_vectors(qcell)
                    nvecs = len(vecs)
                    vecs = self.ensure_float_type(vecs)
                    vecids = self.ensure_vector_id_type(vecids)
                    qcellids = np.zeros(nvecs, dtype=self.FAISS_VECTOR_ID_TYPE)
                    qcellids[:] = qcell
                    # we use 'add_core' which is very fast, since we already know the qcell IDs of the added vectors:
                    index.add_core(nvecs, faiss.swig_ptr(vecs), faiss.swig_ptr(vecids), faiss.swig_ptr(qcellids))
                    loaded_qcells.add(qcell)
        return index
    
    def search_quantizer(self, vectors, k=None):
        """Search the top-'k' qcells for each of 'vectors' in the current quantizer.
        'k' here corresponds to the 'nprobe' parameter used in the FAISS API, that is,
        it controls how many close qcells are exhaustively searched for each query vector.
        """
        k = min(k or self.nprobe, self.nlist)
        vectors = self.ensure_float_type(vectors)
        vectors = self.ensure_vector_array(vectors)
        D, V = self.get_quantizer().search(vectors, k)
        return D, V

    def search(self, vectors, k=1, nprobe=None):
        """Search this index for the 'k' nearest neighbors of 'vectors' and return
        the result as a tuple D, V (where D are distances and V vector IDs).  'nprobe'
        controls how many inverted lists will be searched (defaults to self.nprobe).
        """
        vectors = self.ensure_float_type(vectors)
        vectors = self.ensure_vector_array(vectors)
        _, qcells = self.search_quantizer(vectors, k=nprobe)
        index = self.get_search_index_for_qcells(qcells)
        k = min(k, index.ntotal)
        D, V = index.search(vectors, k)
        return D, V



GRADIENT = 'gradient'
CATEGORICAL = 'categorical'

color_style_dict = {
    'd3.schemeCategory10': {'style': CATEGORICAL, 'length': 10},
    'd3.schemeAccent': {'style': CATEGORICAL, 'length': 8},
    'd3.schemeDark2': {'style': CATEGORICAL, 'length': 8},
    'd3.schemePaired': {'style': CATEGORICAL, 'length': 12},
    'd3.schemePastel1': {'style': CATEGORICAL, 'length': 9},
    'd3.schemePastel2': {'style': CATEGORICAL, 'length': 8},
    'd3.schemeSet1': {'style': CATEGORICAL, 'length': 9},
    'd3.schemeSet2': {'style': CATEGORICAL, 'length': 8},
    'd3.schemeSet3': {'style': CATEGORICAL, 'length': 12},
    'd3.schemeTableau10': {'style': CATEGORICAL, 'length': 10},
    'd3.interpolateBrBG': {'style': GRADIENT},
    'd3.interpolatePRGn': {'style': GRADIENT},
    'd3.interpolatePiYG': {'style': GRADIENT},
    'd3.interpolatePuOr': {'style': GRADIENT},
    'd3.interpolateRdBu': {'style': GRADIENT},
    'd3.interpolateRdGy': {'style': GRADIENT},
    'd3.interpolateRdYlBu': {'style': GRADIENT},
    'd3.interpolateRdYlGn': {'style': GRADIENT},
    'd3.interpolateSpectral': {'style': GRADIENT},
    'd3.interpolateBlues': {'style': GRADIENT},
    'd3.interpolateGreens': {'style': GRADIENT},
    'd3.interpolateOranges': {'style': GRADIENT},
    'd3.interpolatePurples': {'style': GRADIENT},
    'd3.interpolateReds': {'style': GRADIENT},
    'd3.interpolateGreys': {'style': GRADIENT},
    'd3.interpolateTurbo': {'style': GRADIENT},
    'd3.interpolateViridis': {'style': GRADIENT},
    'd3.interpolateInferno': {'style': GRADIENT},
    'd3.interpolateMagma': {'style': GRADIENT},
    'd3.interpolatePlasma': {'style': GRADIENT},
    'd3.interpolateCividis': {'style': GRADIENT},
    'd3.interpolateWarm': {'style': GRADIENT},
    'd3.interpolateCool': {'style': GRADIENT},
    'd3.interpolateCubehelixDefault': {'style': GRADIENT},
    'd3.interpolateBuGn': {'style': GRADIENT},
    'd3.interpolateBuPu': {'style': GRADIENT},
    'd3.interpolateGnBu': {'style': GRADIENT},
    'd3.interpolateOrRd': {'style': GRADIENT},
    'd3.interpolatePuBuGn': {'style': GRADIENT},
    'd3.interpolatePuBu': {'style': GRADIENT},
    'd3.interpolatePuRd': {'style': GRADIENT},
    'd3.interpolateRdPu': {'style': GRADIENT},
    'd3.interpolateYlGnBu': {'style': GRADIENT},
    'd3.interpolateYlGn': {'style': GRADIENT},
    'd3.interpolateYlOrBr': {'style': GRADIENT},
    'd3.interpolateYlOrRd': {'style': GRADIENT},
    'd3.interpolateRainbow': {'style': GRADIENT}
}



"""
Convert a KGTK file to a GitHub markdown table.

--mode=NONE is default.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Convert a KGTK file to a GitHub Markdown Table.',
        'description': 'Convert a KGTK input file to a GitHub markdown table on output. ' +
        '\n\nUse this command to filter the output of any KGTK command: ' +
        '\n\nkgtk xxx / md ' +
        '\n\nUse it to convert a KGTK file to a GitHub Markdown table in a file: ' +
        '\n\nkgtk md -i file.tsv -o file.md' +
        '\n\nThis command defaults to --mode=NONE so it will work with TSV files that do not follow KGTK column naming conventions.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert md --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(who="The KGTK file to convert to a GitHub markdown table.", positional=True)
    parser.add_output_file(who="The GitHub markdown file to write.")

    parser.add_argument(      "--output-format", dest="output_format", help=h("The file format (default=%(default)s)"), type=str, default="md")

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode.NONE,
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        output_format: str,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys
    import typing
    
    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.join.kgtkcat import KgtkCat
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_file_path: Path = KGTKArgumentParser.get_input_file(input_file)
    output_file_path: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # TODO: check that at most one input file is stdin?

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, mode=KgtkReaderMode.NONE)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_file_path), file=error_file, flush=True)
        print("--output-file=%s" % str(output_file_path), file=error_file, flush=True)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        kc: KgtkCat = KgtkCat(input_file_paths=[input_file_path],
                              output_path=output_file_path,
                              output_format=output_format,
                              reader_options=reader_options,
                              value_options=value_options,
                              error_file=error_file,
                              verbose=verbose,
                              very_verbose=very_verbose
        )
        
        kc.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




import unittest
from kgtk.cli_entry import cli_entry
from kgtk.cli.dummy import run
from kgtk.exceptions import KGTKException


class TestDummy(unittest.TestCase):
    def test_module(self):
        # test separate module files
        pass

    def test_run(self):
        # test run function
        # exceptions here are not trapped by KGTKExceptionHandler
        with self.assertRaises(KGTKException):
            run(name="kgtk", info=None, error=True, _debug=False)

    def test_cli(self):
        # test command from cli entry
        assert cli_entry("kgtk", "dummy", "normal_test") == 0
        assert cli_entry("kgtk", "dummy", "test_exception", "-e") != 0


if __name__ == "__main__":
    unittest.main()
